# -*- coding: utf-8 -*-
"""

    Инструменты для проверки значений

"""
# встроенные модули
from typing import Collection, Optional, List, TypeVar, Any

T = TypeVar('T')
OptCol = Optional[Collection[T]]


def filter_by_inclusion(collection: Collection[T], only: OptCol = None,
                        without: OptCol = None, sort: bool = True) -> List[T]:
    """
    Отфильтровать по вхождению в диапазон
    """
    result = list(collection)

    if only:
        only = set(only)
        result = [x for x in result if x in only]

    if without:
        without = set(without)
        result = [x for x in result if x not in without]

    if sort:
        return sorted(result)
    return result


def filter_by_value(collection: Collection[T], start: Optional[Any] = None,
                    stop: Optional[Any] = None, sort: bool = True) -> List[T]:
    """
    Отфильтровать по величине
    """
    result = list(collection)

    if start:
        result = [x for x in result if x >= start]

    if stop:
        result = [x for x in result if x <= stop]

    if sort:
        return sorted(result)
    return result



# -*- coding: utf-8 -*-
"""

    Базовый класс для управления конфигурациями

"""
# встроенные модули
import os
import json
from abc import abstractmethod, ABC

# модули проекта
from trivial_tools.special.special import fail


class AbstractConfig(ABC):
    """
    Абстрактный класс для хранения параметров
    """
    default_config_name: str = 'base'

    @abstractmethod
    def __init__(self, **kwargs):
        """
        Инициация
        """

    @staticmethod
    def get_config_path(filename: str) -> str:
        """
        Получить путь к файлу конфигурации
        """
        folder = os.path.abspath(os.getcwd())
        if folder.endswith('tests'):
            # при тестировании конфиг обычно находится на каталог выше
            parts = os.path.split(folder)
            return os.path.join(*parts[:-1], filename)
        return os.path.join(folder, filename)

    @classmethod
    def extract_json(cls, config_name: str, path: str) -> dict:
        """
        Вытащить данные из JSON файла
        """
        if not os.path.exists(path):
            fail(f'Не найден файл конфигурации: "{path}"', reason=FileNotFoundError)

        with open(path, mode='r', encoding='utf-8') as file:
            data = json.load(file)

        if config_name not in data:
            fail(f'В конфиге не представлен раздел "{config_name}"', reason=KeyError)

        config = {**data.get(cls.default_config_name, {}), **data[config_name]}

        if config:
            return config

        fail(f'Не удалось загрузить конфигурацию "{config_name}" из файла: "{path}"',
             reason=KeyError)

    @classmethod
    def from_json(cls, config_name: str, filename: str = 'config.json'):
        """
        Собрать экземпляр из JSON файла
        """
        path = cls.get_config_path(filename)
        data = cls.extract_json(config_name, path)
        return cls(created_from=path, **data)



# -*- coding: utf-8 -*-
"""

    Особые инструменты

"""
# встроенные модули
from itertools import zip_longest
from typing import NoReturn, Optional, Type, Sequence, Callable, List, Any

# модули проекта
from trivial_tools.formatters.base import decorate


def fail(message: str, reason: Type[Exception] = RuntimeError,
         raise_from: Optional[Exception] = None) -> NoReturn:
    """Выбросить исключение с заданным текстовым сообщением

    :param message: сообщение, которое необходимо передать
    :param reason: исключение, которые необходимо вызвать
    :param raise_from: возбудить исключение из переданного экземпляра
    :return: нет возврата
    """
    if isinstance(reason, (UnicodeDecodeError, UnicodeTranslateError, UnicodeEncodeError)):
        # Unicode ошибки нельзя выбросить просто с одним текстовым сообщением
        raise

    if isinstance(raise_from, Exception):
        # выбрасываем из экземпляра класса исключений
        name = raise_from.__class__.__name__
        instance = type(raise_from)(decorate(f'{name}: {message}'))
        raise instance from raise_from

    # выбрасываем из класса исключений
    name = reason.__name__

    # В настоящий момент (python 3.7) нет способа корректно указать исключения в аннотациях типов.
    # noinspection PyCallingNonCallable
    raise reason(decorate(f'{name}: {message}'))


def group_cast(elements: Sequence, *casters: Callable) -> List[Any]:
    """Выполнить групповое преобразование типа.

    Может сэкономить немного места на конверсиях.

    :param elements: последовательность элементов для преобразования.
    :param casters: последовательность типов или функций для преобразования
    :return: список преобразованных элементов
    """
    if not casters:
        return list(elements)

    elif casters[-1] != Ellipsis and len(casters) == len(elements):
        return [func(val) for func, val in zip(casters, elements)]

    elif casters[-1] == Ellipsis and len(casters) >= 2:
        return [func(val) for func, val in
                zip_longest(casters[:-1], elements, fillvalue=casters[-2])]

    fail(f'Количество преобразователей ({len(casters)}) не '
         f'совпадает с количеством аргументов ({len(elements)})!', reason=ValueError)



# -*- coding: utf-8 -*-
"""

    Тесты специфических операций

"""
# сторонние модули
import pytest

# модули проекта
from trivial_tools.special.special import group_cast


def test_group_cast():
    """
    Должен преобразовывать типы
    """
    raw = ['1', '2', '3']
    assert group_cast(raw) == raw
    assert group_cast(raw, int, float, str) == [1, 2.0, '3']


def test_group_cast_ellipsis():
    """
    Должен преобразовывать типы для больших коллекий
    """
    raw = ['1', '2', '3']
    assert group_cast(raw) == raw
    assert group_cast(raw, int, ...) == [1, 2, 3]
    assert group_cast(raw, str, int, ...) == ['1', 2, 3]


def test_group_cast_fail():
    """
    Должен преобразовывать типы
    """
    with pytest.raises(ValueError):
        group_cast([1, 2], float, float, float)

    with pytest.raises(ValueError):
        group_cast([1, 2], float)



# -*- coding: utf-8 -*-
"""

    Инструменты проверок типов

"""
# встроенные модули
from typing import Any


def is_int(something: Any) -> bool:
    """Проверить, можно ли этот объект преобразовать в int.

    :param something: неизвестный объект, который надо преобразовать
    :return: True если можно преобразовать, False если нельзя

    >>> is_int('25')
    True

    >>> is_int(25.4)
    True

    >>> is_int('0.25')
    False
    """
    try:
        int(something)
        return True
    except (ValueError, TypeError):
        # Для вариантов типа str(0.25) базовая проверка возвращает False!
        return False


def is_float(something: Any) -> bool:
    """Проверить, можно ли этот объект преобразовать в float.

    :param something: неизвестный объект, который надо преобразовать
    :return: True если можно преобразовать, False если нельзя

    >>> is_float('0.25')
    True

    >>> is_float(25)
    True

    >>> is_float('test')
    False
    """
    try:
        float(something)
        return True
    except (ValueError, TypeError):
        return False



# -*- coding: utf-8 -*-
"""

    Инструменты форматирования текста

"""
# встроенные модули
from typing import Any, Sequence, Union


def decorate(message: str, symbol: str = '+', line_width: int = 79) -> str:
    """Вывести текст в обрамлении символов

    :param message: оригинальное сообщение
    :param symbol: каким символом написать строки
    :param line_width: какую ширину заполнить символами
    :return: оформленное текстовое сообщение в виде нескольких строк
    """
    separator = '\n' + symbol * line_width + '\n'
    result = separator + message.strip() + separator
    return result


def s_type(something: Any) -> str:
    """
    Вывести укороченное название типа

    >>> s_type(9)
    'int'

    >>> s_type(None)
    'NoneType'

    >>> class MyTestClass: pass
    >>> s_type(MyTestClass())
    'MyTestClass'

    >>> s_type(Union[int, type(None)])
    'typing.Union[int, NoneType]'
    """
    if isinstance(something, type):
        return something.__name__

    result = type(something).__name__
    if result == '_GenericAlias':
        return str(something).lstrip('typing.')

    return result


def shorten(something: Any) -> str:
    """
    Выдать укороченную форму записи
    """
    if isinstance(something, str):
        if len(something) <= 20:
            result = repr(something)
        else:
            result = repr(something[:10] + ' ... ' + something[-10:])

    elif isinstance(something, Sequence):
        if len(something) <= 10:
            result = repr(something)
        else:
            result = (', '.join(repr(x) for x in something[0:4])
                      + ' ... '
                      + ', '.join(repr(x) for x in something[-4:]))

            if isinstance(something, list):
                result = f'[{result}]'

            elif isinstance(something, tuple):
                result = f'({result})'
    else:
        result = repr(something)

    return result


def detail(something: Any) -> str:
    """
    Вывести параметри неизвестного объекта

    >>> detail(9)
    '9, int'
    """
    return f'{shorten(something)}, {s_type(something)}'


def dict_as_args(input_dict: dict) -> str:
    """
    Разложить словарь на последовательность аргументов будто это ключевые слова

    >>> dict_as_args(dict(a=1, b=2, c="test"))
    "a=1, b=2, c='test'"
    """
    return ', '.join(f'{key}={value!r}' for key, value in input_dict.items())



# -*- coding: utf-8 -*-
"""

    Тесты скользящего среднего

"""
# сторонние модули
import pytest
from pytest import approx

# модули проекта
from trivial_tools.containers.class_moving_average import MovingAverage


@pytest.fixture()
def numbers():
    """
    Исходные данные для вычисления среднего
    """
    return [1, 3, 5, 6, 8, 41, 9, 6, 32, 58, 8, 78, 3, 2, 45, 6, 9, 87, 51, 23]


@pytest.fixture()
def reference():
    """
    Эталонные средние
    """
    return [1, 2, 3, 3.75, 5.5, 15, 16, 16, 22, 26.25, 26, 44,
            36.75, 22.75, 32, 14, 15.5, 36.75, 38.25, 42.5]


@pytest.fixture()
def summary():
    """
    Эталонные суммы
    """
    return [1, 4, 9, 15, 22, 60, 64, 64, 88, 105, 104, 176, 147, 91, 128, 56, 62, 147, 153, 170]


def test_creation():
    """
    Проверка создания
    """
    with pytest.raises(ValueError):
        MovingAverage()

    m = MovingAverage(window=2)
    assert approx(m.avg) == 0.0
    assert approx(m.sum) == 0.0

    m = MovingAverage([1, 2, 3])
    assert approx(m.avg) == 2.0
    assert approx(m.sum) == 6.0

    m = MovingAverage([1, 2, 3], window=2)
    assert approx(m.avg) == 2.5
    assert approx(m.sum) == 5.0


def test_pushing(reference, summary, numbers):
    """
    Проверка добавления элементов
    """
    m = MovingAverage(window=4)

    for ref, summ, number in zip(reference, summary, numbers):
        m.push(number)
        assert approx(m.avg) == ref
        assert approx(m.sum) == summ


def test_enlarge(numbers):
    """
    Проверка увеличения размеров
    """
    m = MovingAverage(window=4)
    for number in numbers:
        m.push(number)

    assert approx(m.avg) == 42.5
    assert approx(m.sum) == 170.0

    m.resize(10)

    assert approx(m.avg) == 42.5
    assert approx(m.sum) == 170.0

    reference_large = [34.2, 29, 25.571428, 23.125, 21.444444, 23.4, 23.4, 15.3, 13.4, 16.9,
                       17.6, 25.1, 24.9, 24.5, 28.2, 24.7, 24.7, 32.8, 34.7, 31.2]

    summary_large = [171, 174, 179, 185, 193, 234, 234, 153, 134, 169,
                     176, 251, 249, 245, 282, 247, 247, 328, 347, 312]

    for ref, summ, number in zip(reference_large, summary_large, numbers):
        m.push(number)
        assert approx(m.avg) == ref
        assert approx(m.sum) == summ


def test_shrink(numbers):
    """
    Проверка уменьшения размеров
    """
    m = MovingAverage(window=4)
    for number in numbers:
        m.push(number)

    assert approx(m.avg) == 42.5
    assert approx(m.sum) == 170.0

    m.resize(3)

    assert approx(m.avg) == 53.666666
    assert approx(m.sum) == 161.0

    reference_small = [25, 9, 3, 4.666666, 6.333333, 18.33333, 19.33333, 18.66666, 15.66666, 32,
                       32.66666, 48, 29.66666, 27.66666, 16.66666, 17.66666, 20, 34, 49, 53.66666]

    summary_small = [75, 27, 9, 14, 19, 55, 58, 56, 47, 96, 98,
                     144, 89, 83, 50, 53, 60, 102, 147, 161]

    for ref, summ, number in zip(reference_small, summary_small, numbers):
        m.push(number)
        assert approx(m.avg) == ref
        assert approx(m.sum) == summ


def test_getitem(numbers):
    """
    Проверка обращения к элементу
    """
    m = MovingAverage(window=4)
    for number in numbers:
        m.push(number)

    assert approx(m[0]) == 9.0
    assert approx(m[1]) == 87.0
    assert approx(m[2]) == 51.0
    assert approx(m[3]) == 23.0

    assert m[:] == [9.0, 87.0, 51.0, 23.0]

    assert approx(m.avg) == 42.5
    assert approx(m.sum) == 170.0


def test_setitem(numbers):
    """
    Проверка подмены элемента
    """
    m = MovingAverage(window=4)
    for number in numbers:
        m.push(number)

    m[0] = 99.0

    assert approx(m[0]) == 99.0
    assert approx(m[1]) == 87.0
    assert approx(m[2]) == 51.0
    assert approx(m[3]) == 23.0

    assert approx(m.avg) == 65.0
    assert approx(m.sum) == 260.0

    with pytest.raises(IndexError):
        m[:] = [7.0, 28.0, 31.0, -900.0]



# -*- coding: utf-8 -*-
"""

    Метакласс для реализации паттерна одиночка

"""
# модули проекта
from trivial_tools.special.special import fail
from trivial_tools.formatters.base import s_type


class MetaSingleton(type):
    """
    Метакласс для реализации паттерна одиночка
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        """
        Вызов может быть от различных классов!
        """
        if cls not in MetaSingleton._instances:
            MetaSingleton._instances[cls] = super(MetaSingleton, cls).__call__(*args, **kwargs)
        return MetaSingleton._instances[cls]

    @classmethod
    def get_instance(mcs, required_cls):
        """
        Обращение от потомка за экземпляром себя
        """
        if isinstance(required_cls, type):
            if required_cls in mcs._instances:
                return mcs._instances[required_cls]
        else:
            if type(required_cls) in mcs._instances:
                return mcs._instances[type(required_cls)]

        fail(f'Метакласс {mcs} не содержит экземпляров {s_type(required_cls)}!', reason=KeyError)



# -*- coding: utf-8 -*-
"""

    Работа с распечаткой из pandas

"""
# сторонние модули
import pandas as pd


def pandas_output(*args, min_rows: int = 100, max_rows: int = 100, max_columns: int = 100,
                  expand_frame_repr: bool = False):
    """
    Распечатать с настройками контекста
    """
    with pd.option_context(
            'display.min_rows', min_rows,
            'display.max_rows', max_rows,
            'display.max_columns', max_columns,
            'display.expand_frame_repr', expand_frame_repr
    ):
        print(*args)



# -*- coding: utf-8 -*-
"""

    Тесты описания прошедшего времени

"""
# встроенные модули
from datetime import timedelta

# модули проекта
from trivial_tools.datetime_tools.passed_time import form_minutes, form_weeks, form_days
from trivial_tools.datetime_tools.passed_time import form_seconds, form_hours, passed_time


def test_weeks():
    """
    Проверить отображение недель
    """
    assert form_weeks(timedelta(days=0)) == (None, 0)
    assert form_weeks(timedelta(days=2)) == (None, 2)
    assert form_weeks(timedelta(days=5)) == (None, 5)
    assert form_weeks(timedelta(days=7)) == ('1 неделя', 0)
    assert form_weeks(timedelta(days=10)) == ('1 неделя', 3)
    assert form_weeks(timedelta(days=14)) == ('2 недели', 0)
    assert form_weeks(timedelta(days=21)) == ('3 недели', 0)
    assert form_weeks(timedelta(days=28)) == ('4 недели', 0)
    assert form_weeks(timedelta(days=35)) == ('5 недель', 0)
    assert form_weeks(timedelta(days=42)) == ('6 недель', 0)


def test_days():
    """
    Проверить отображение дней
    """
    assert form_days(timedelta(days=0)) is None
    assert form_days(timedelta(days=1)) == '1 день'
    assert form_days(timedelta(days=2)) == '2 дня'
    assert form_days(timedelta(days=3)) == '3 дня'
    assert form_days(timedelta(days=4)) == '4 дня'
    assert form_days(timedelta(days=5)) == '5 дней'
    assert form_days(timedelta(days=6)) == '6 дней'


def test_hours():
    """
    Проверить отображение часов
    """
    assert form_hours(timedelta(seconds=0)) == (None, 0)
    assert form_hours(timedelta(seconds=15)) == (None, 15)
    assert form_hours(timedelta(minutes=1)) == (None, 60)
    assert form_hours(timedelta(minutes=30)) == (None, 1800)
    assert form_hours(timedelta(minutes=59)) == (None, 3540)
    assert form_hours(timedelta(hours=0)) == (None, 0)
    assert form_hours(timedelta(hours=1)) == ('1 час', 0)
    assert form_hours(timedelta(hours=3)) == ('3 часа', 0)
    assert form_hours(timedelta(hours=7)) == ('7 часов', 0)
    assert form_hours(timedelta(hours=10)) == ('10 часов', 0)
    assert form_hours(timedelta(hours=15)) == ('15 часов', 0)
    assert form_hours(timedelta(hours=23)) == ('23 часа', 0)


def test_minutes():
    """
    Проверить отображение минут
    """
    assert form_minutes(timedelta(seconds=0)) == (None, 0)
    assert form_minutes(timedelta(seconds=15)) == (None, 15)
    assert form_minutes(timedelta(minutes=1)) == ('1 минута', 0)
    assert form_minutes(timedelta(minutes=2)) == ('2 минуты', 0)
    assert form_minutes(timedelta(minutes=5)) == ('5 минут', 0)
    assert form_minutes(timedelta(minutes=7)) == ('7 минут', 0)
    assert form_minutes(timedelta(minutes=10)) == ('10 минут', 0)
    assert form_minutes(timedelta(minutes=11)) == ('11 минут', 0)
    assert form_minutes(timedelta(minutes=12)) == ('12 минут', 0)
    assert form_minutes(timedelta(minutes=13)) == ('13 минут', 0)
    assert form_minutes(timedelta(minutes=14)) == ('14 минут', 0)
    assert form_minutes(timedelta(minutes=15)) == ('15 минут', 0)
    assert form_minutes(timedelta(minutes=16)) == ('16 минут', 0)
    assert form_minutes(timedelta(minutes=17)) == ('17 минут', 0)
    assert form_minutes(timedelta(minutes=19)) == ('19 минут', 0)
    assert form_minutes(timedelta(minutes=21)) == ('21 минута', 0)
    assert form_minutes(timedelta(minutes=46)) == ('46 минут', 0)
    assert form_minutes(timedelta(minutes=53)) == ('53 минуты', 0)
    assert form_minutes(timedelta(minutes=59)) == ('59 минут', 0)


def test_seconds():
    """
    Проверить отображение секунд
    """
    assert form_seconds(timedelta(seconds=0)) is None
    assert form_seconds(timedelta(seconds=1)) == '1 секунда'
    assert form_seconds(timedelta(seconds=2)) == '2 секунды'
    assert form_seconds(timedelta(seconds=3)) == '3 секунды'
    assert form_seconds(timedelta(seconds=5)) == '5 секунд'
    assert form_seconds(timedelta(seconds=7)) == '7 секунд'
    assert form_seconds(timedelta(seconds=10)) == '10 секунд'
    assert form_seconds(timedelta(seconds=11)) == '11 секунд'
    assert form_seconds(timedelta(seconds=12)) == '12 секунд'
    assert form_seconds(timedelta(seconds=13)) == '13 секунд'
    assert form_seconds(timedelta(seconds=17)) == '17 секунд'
    assert form_seconds(timedelta(seconds=19)) == '19 секунд'
    assert form_seconds(timedelta(seconds=20)) == '20 секунд'
    assert form_seconds(timedelta(seconds=21)) == '21 секунда'
    assert form_seconds(timedelta(seconds=22)) == '22 секунды'
    assert form_seconds(timedelta(seconds=23)) == '23 секунды'
    assert form_seconds(timedelta(seconds=24)) == '24 секунды'
    assert form_seconds(timedelta(seconds=25)) == '25 секунд'
    assert form_seconds(timedelta(seconds=26)) == '26 секунд'
    assert form_seconds(timedelta(seconds=30)) == '30 секунд'
    assert form_seconds(timedelta(seconds=31)) == '31 секунда'
    assert form_seconds(timedelta(seconds=32)) == '32 секунды'
    assert form_seconds(timedelta(seconds=33)) == '33 секунды'
    assert form_seconds(timedelta(seconds=46)) == '46 секунд'
    assert form_seconds(timedelta(seconds=51)) == '51 секунда'
    assert form_seconds(timedelta(seconds=53)) == '53 секунды'
    assert form_seconds(timedelta(seconds=59)) == '59 секунд'


def test_combined():
    """
    Общая проверка работы всех блоков
    """
    delta = timedelta(seconds=0)
    assert passed_time(delta) == '0 секунд'

    delta = timedelta(weeks=4, days=6, hours=22, minutes=47, seconds=17)
    assert passed_time(delta) == '4 недели, 6 дней, 22 часа, 47 минут, 17 секунд'

    delta = timedelta(seconds=3601)
    assert passed_time(delta) == '1 час, 1 секунда'

    delta = timedelta(days=33)
    assert passed_time(delta) == '4 недели, 5 дней'

    delta = timedelta(seconds=53601)
    assert passed_time(delta) == '14 часов, 53 минуты, 21 секунда'

    delta = timedelta(hours=25, seconds=86)
    assert passed_time(delta) == '1 день, 1 час, 1 минута, 26 секунд'



# -*- coding: utf-8 -*-
"""

    Получение информации о том, в каком коммите мы сейчас находимся

"""
# встроенные модули
import subprocess


def get_git_revision_hash(cmd: str = 'git rev-parse HEAD',
                          encoding: str = 'utf-8') -> str:
    """Получить хеш текущего коммита.

    Обращается к системе и получает данные о том,
    какой коммит сейчас является головным для данного репозитория. Исполнение зависит от
    параметров системы и настроек окружения.

    :param cmd: команда для исполнения подпроцессом

    :param encoding: из какой кодировки произвести конвертацию
    (стандартная кодировка терминала на Windows - cp866)

    Пример вывода:
        'aa4d5ef11973828a18425c369e0fdde5fd45c11d'
    """
    pipe = subprocess.Popen(
        cmd,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        shell=True
    )
    stdout, stderr = pipe.communicate()
    response = stdout.decode(encoding)

    if response.replace('"', '').replace("'", '').startswith(('git', 'fatal')):
        response = 'unknown'
    else:
        # есть риск получить сюда очень большой блок текста в виде
        # стандартного ответа гита на ввод незнакомой команды
        response = response[:41].strip()

    return response


def get_git_revision_hash_short(cmd: str = 'git rev-parse --short HEAD',
                                encoding: str = 'utf-8') -> str:
    """Получить укороченный хеш текущего коммита.

    :param cmd: команда для исполнения подпроцессом

    :param encoding: из какой кодировки произвести конвертацию
    (стандартная кодировка терминала на Windows - cp866)

    Пример вывода:
        'aa4d5ef'
    """
    return get_git_revision_hash(cmd, encoding)



# -*- coding: utf-8 -*-
"""

    Двусторонная карусель.
    В работе подобна deque, но не меняет размера при добавлении и извлечении элементов.

"""
# встроенные модули
from typing import Any, Optional, Collection, Generator, List

# модули проекта
from trivial_tools.special.special import fail
from trivial_tools.formatters.base import s_type
from trivial_tools.containers.class_carousel import Carousel


class DeCarousel(Carousel):
    """
    Двусторонная карусель.
    В работе подобна deque, но не меняет размера при добавлении и извлечении элементов.
    """
    __slots__ = ('_head', '_tail')

    def __init__(self, source: Optional[Collection] = None,
                 window: int = 0, sentinel: Any = object()):
        """
        Создание экземпляра. Можно задать базовую коллекцию, размер и объект для пустых слотов

        :param source: исходная коллекция элементов, на базе которой надо собрать экземпляр
        :param window: максимальное количество элементов (ширина окна вычисления)
        :param sentinel: элемент для заполнения пустых ячеек (можно добавить свой)
        """
        self._head = -1
        self._tail = -1
        super().__init__(source, window, sentinel)

    def __contains__(self, item: Any) -> bool:
        """
        Проверка наличия элемента
        """
        return item in self._data

    def _step_right(self, index: int) -> int:
        """
        Шагнуть вправо
        """
        if index < self.window - 1:
            index += 1
        else:
            index = 0
        return index

    def _step_left(self, index: int) -> int:
        """
        Шагнуть влево
        """
        if index > 0:
            index -= 1
        else:
            index = self.window - 1
        return index

    @property
    def left(self) -> Optional[Any]:
        """
        Крайний элемент слева
        """
        if len(self):
            return self._data[self._head]
        return None

    @property
    def right(self) -> Optional[Any]:
        """
        Крайний элемент справа
        """
        if len(self):
            return self._data[self._tail]
        return None

    def restore(self) -> None:
        """
        Сбросить индекс и длину, заполнить
        внутреннее хранилище пустыми объектами
        """
        self._len = 0
        self._index = 0
        self._head = -1
        self._tail = -1
        self._data = [self._sentinel for _ in range(self.window)]

    def _internals(self) -> Generator[Any, None, None]:
        """
        Проитерироваться по элементам внутреннего хранилища (включая пустые элементы!)
        """
        if self._head == -1 and self._tail == -1:
            yield from self._data

        else:
            i = self._head
            while True:
                if i == self._tail:
                    break
                yield self._data[i]
                i = self._step_right(i)
            yield self._data[self._tail]

    def update_len(self) -> None:
        """
        Пересчитать длину коллекции
        """
        if self._head < self._tail:
            self._len = self._tail - self._head + 1

        elif self._head > self._tail:
            self._len = self._tail - self._head + 1 + self.window

        elif self._head == self._tail:
            self._len = 1

    def push(self, item: Any) -> Any:
        """
        Добавить элемент в карусель

        Мы замещаем элемент, выталкивая его из списка и возвращаем его. Его значение потом может
        быть использовано на вызывающей стороне

        :param item: Новый элемент любого типа
        :return: Старый элемент, который был вытолкнут при добавлении (может оказаться sentinel!)
        """
        old_value = self._data[self._index]
        self._data[self._index] = item

        if self._head == self._tail == -1:
            # первая вставка элемента
            self._index = self._step_right(self._index)
            self._head = self._step_right(self._head)
            self._tail = self._step_right(self._tail)

        else:
            # обычная конфигурация
            if self._index == self._head:
                # затираем голову
                self._head = self._step_right(self._head)

            self._index = self._step_right(self._index)
            self._tail = self._step_right(self._tail)

        self.update_len()
        return old_value

    def append(self, item: Any) -> None:
        """
        Добавить элемент справа
        """
        self.push(item)
        return None

    def pop(self) -> Any:
        """
        Извлечь элемент справа
        """
        if self._len == 0:
            fail(f'Попытка извлечь элемент из пустой коллекции: {self}', reason=IndexError)

        value = self._data[self._tail]
        self._data[self._tail] = self._sentinel
        self._tail = self._step_left(self._tail)
        self._index = self._step_left(self._index)
        self._len -= 1
        return value

    def appendleft(self, item: Any) -> None:
        """
        Добавить элемент слева
        """
        if self._head == self._tail == -1:
            # первая вставка элемента
            self._data[-1] = item
            self._index = 0
            self._head = len(self._data) - 1
            self._tail = len(self._data) - 1

        else:
            # обычная конфигурация
            self._head = self._step_left(self._head)
            if self._head == self._tail:
                self._tail = self._step_left(self._tail)

            self._data[self._head] = item

        self.update_len()

    def popleft(self) -> Any:
        """
        Извлечь элемент слева
        """
        if self._len == 0:
            fail(f'Попытка извлечь элемент из пустой коллекции: {self}', reason=IndexError)

        value = self._data[self._head]
        self._data[self._head] = self._sentinel
        self._head = self._step_right(self._head)
        self._index = self._step_right(self._tail)
        self._len -= 1

        if self._len == 0:
            self.restore()
        return value



# -*- coding: utf-8 -*-
"""

    Дополнительные тесты двусторонней карусели

"""
# сторонние модули
import pytest

# модули проекта
from trivial_tools.containers.class_de_carousel import DeCarousel


def test_full_cycle():
    """
    Проверка работы
    """
    d = DeCarousel(window=4, sentinel=None)

    for i in range(100):
        d.append(i)

    assert d.get_contents() == [96, 97, 98, 99]
    assert d.left == 96
    assert d.right == 99
    assert len(d) == 4

    for i in range(3):
        d.popleft()

    assert d.get_contents() == [99]
    assert d._data == [None, None, None, 99]
    assert d._index == 0
    assert d.left == 99
    assert d.right == 99
    assert len(d) == 1

    d.append(1)
    assert d.get_contents() == [99, 1]
    assert d._data == [1, None, None, 99]
    assert d.left == 99
    assert d.right == 1
    assert len(d) == 2

    for i in range(1, 200):
        d.append(i)

    assert d.get_contents() == [196, 197, 198, 199]
    assert d._data == [196, 197, 198, 199]
    assert d.left == 196
    assert d.right == 199
    assert len(d) == 4



# -*- coding: utf-8 -*-
"""

    Тесты класса хранящего экземпляр себя

"""
# сторонние модули
import pytest

# модули проекта
from trivial_tools.classes.conservator import Conservator


@pytest.fixture
def class_():
    class Test(Conservator):
        def __init__(self, var):
            self.var = var
    return Test


def test_creation(class_):
    """
    Создание экземпляра
    """
    t = class_(1)
    assert t.var == 1


def test_double_creation(class_):
    """
    Создание экземпляра
    """
    t = class_(1)
    assert t.var == 1


def test_register(class_):
    """
    Добавить валидный экземпляр
    """
    assert not class_.has_instance()
    inst_1 = class_(1)
    assert not class_.has_instance()
    class_.register(inst_1)

    inst_2 = class_(2)
    inst_3 = class_(3)

    assert class_.get_instance() is inst_1

    class_.unregister()
    class_.register(inst_2)
    assert class_.get_instance() is inst_2

    class_.replace(inst_3)
    assert class_.get_instance() is inst_3


def test_register_force(class_):
    """
    Добавить валидный экземпляр
    """
    assert not class_.has_instance()

    with pytest.raises(ValueError):
        class_.register(1)

    class_.register_force(256)

    assert class_.get_instance() == 256
    assert class_.has_instance()

    with pytest.raises(ValueError):
        inst = class_(3)
        class_.register_force(inst)


def test_unregister(class_):
    """
    Удалить валидный экземпляр
    """
    assert not class_.has_instance()
    with pytest.raises(ValueError):
        class_.unregister()

    t = class_(1)
    class_.register(t)
    t.unregister()
    assert not class_.has_instance()


def test_replace(class_):
    """
    Заменить валидный экземпляр
    """
    assert not class_.has_instance()
    inst = class_(1)
    class_.register(inst)
    assert class_.has_instance()

    with pytest.raises(ValueError):
        class_.replace(25)

    with pytest.raises(ValueError):
        class_.register(25)

    assert class_.has_instance()
    assert class_.get_instance() is inst

    class_.unregister()
    assert not class_.has_instance()

    class_.register_force(25)

    assert class_.has_instance()
    assert class_.get_instance() == 25


def test_get_instance(class_):
    """
    Получить валидный экземпляр
    """
    assert not class_.has_instance()

    with pytest.raises(ValueError):
        class_.get_instance()



# -*- coding: utf-8 -*-
"""

    Кеширующий класс

"""
# встроенные модули
from types import FunctionType
from typing import Any, Dict, Optional

# модули проекта
from trivial_tools.storage.caching_instance import CachingInstance


class CachingMachine:
    """
    Специальный класс для выполнения функции локального кеша.
    Redis для бедных. Подразумевается, что эта штука может его
    временно заменить при сильной необходимости
    """
    def __init__(self, expiration: Optional[int] = None, max_items: int = 1000):
        self.expiration = expiration
        self.max_items = max_items
        self._cache: Dict[str, CachingInstance] = {}

    def __getitem__(self, item):
        """
        Прямой доступ к содержимому словаря
        """
        return self._cache[item]

    def __contains__(self, item):
        """
        Проверка наличия внутри кеша
        """
        return item in self._cache

    def extract(self, key: Any) -> Optional[Any]:
        """
        Общая функция извлечения значения. В отличии от редиса,
        мы можем хранить какие попало значения, поэтому в итоге всё сводится к
        одной этой функции
        """
        if key in self._cache:
            instance = self._cache[key]
            if instance.not_expired():
                return self._cache[key].value
            else:
                self.delete(key)
        return None

    def assign(self, key: Any, value: Any, expires: Optional[int] = None) -> bool:
        """
        Общая функция установки значения. В отличии от редиса,
        мы можем хранить какие попало значения, поэтому в итоге всё сводится к
        одной этой функции. Типы должны быть проверены перед вызовом
        """
        if self.is_full():
            self.cleanup()
            if self.is_full():
                return False

        if expires is None:
            expires = self.expiration

        self._cache[key] = CachingInstance(
            value=value,
            expires=expires
        )
        return True

    def get(self, key: str) -> Optional[str]:
        """
        Получить строковое значение по ключу
        """
        return self.extract(key)

    def set(self, key: str, value: str, expires: Optional[int] = None) -> bool:
        """
        Установить строковое значение по ключу
        """
        if not isinstance(value, str):
            print('Метод set принимает только строковые значения!')
            return False

        return self.assign(key, value, expires)

    def exists(self, key: str) -> bool:
        """
        Проверить, есть ли у нас значение для этого ключа
        """
        return key in self._cache

    def delete(self, key: str) -> bool:
        """
        Удалить значение из кеша
        """
        if self.exists(key):
            del self._cache[key]
            return True
        return False

    def total(self) -> int:
        """
        Узнать сколько элементов мы храним
        """
        return len(self._cache)

    def clear(self):
        """
        Очистить память
        """
        self._cache.clear()

    def cleanup(self) -> int:
        """
        Удалить устаревшие элементы
        """
        removed = 0
        keys = list(self._cache.keys())
        for key in keys:
            if self.exists(key) and self[key].expired():
                self.delete(key)
                removed += 1
        return removed

    def is_full(self):
        """
        Проверить заполнен ли кеш
        """
        return self.total() == self.max_items

    def keys(self) -> list:
        """
        Получить все ключи машины
        """
        return list(self._cache.keys())

    def cache_call(self, expires: Optional[int] = None):
        """
        Кешировать результат выполнения функции
        """
        def decorator(func: FunctionType):
            def wrapper(*args):
                key = (func.__name__, args)
                if key in self:
                    result = self.extract(key)
                else:
                    result = func(*args)
                    self.assign(key, result, expires)
                return result
            return wrapper
        return decorator

    def cache_call_using_strings(self, expires: Optional[int] = None):
        """
        Кешировать результат выполнения функции, но ключи хранить в виде строк
        """
        def decorator(func: FunctionType):
            def wrapper(*args):
                key = f'{func.__name__}_' + '_'.join(str(x) for x in args)
                if key not in self:
                    result = func(*args)
                    self.assign(key, result, expires)
                value = self.extract(key)
                return value
            return wrapper
        return decorator



# -*- coding: utf-8 -*-
"""

    Инструменты обработки JSON файлов

"""
# встроенные модули
import os
import sys
import json
from typing import Dict, Any, Optional

# сторонние модули
from loguru import logger


def json_config_load(filename: str, config_name: Optional[str],
                     default_config: str) -> Dict[str, Any]:
    """
    Открыть указанный файл и прочитать его содержимое
    """
    path = os.path.join(os.getcwd(), filename)

    if not os.path.exists(path):
        logger.critical(f'Не найден файл конфигурации: "{path}"')
        raise FileNotFoundError

    with open(path, mode='r', encoding='utf-8') as file:
        data = json.load(file)

    if default_config not in data:
        logger.critical(f'В конфиге не представлен обязательный раздел "{default_config}"')
        raise KeyError

    if config_name in data:
        resulting_config = {**data.get('base', {}), **data[config_name]}
    else:
        resulting_config = {**data.get('base', {}), **data[default_config]}

    if resulting_config:
        return resulting_config

    logger.critical(f'Не удалось загрузить конфигурацию "{config_name}" из файла: "{path}"')
    sys.exit(1)



# -*- coding: utf-8 -*-
"""

    Тесты кеширующего инстанса

"""
# встроенные модули
from datetime import datetime
from unittest.mock import patch

# модули проекта
from trivial_tools.storage.caching_instance import CachingInstance


def test_expiration_persistent():
    """
    Экземпляр без уставки на протухание
    """
    inst = CachingInstance(123)
    assert inst.value == 123
    assert inst.not_expired
    assert inst.not_expired()
    assert not inst.expired()


def test_expiration():
    """
    Проверка протухания экземпляра
    """
    with patch('trivial_tools.storage.caching_instance.datetime') as fake:
        fake.now.return_value = datetime(2019, 10, 31, 18, 6, 0)
        inst = CachingInstance(123, 5)
        assert inst.value == 123
        assert inst.expires == datetime(2019, 10, 31, 18, 6, 5)
        assert inst.not_expired()
        assert not inst.expired()

    with patch('trivial_tools.storage.caching_instance.datetime') as fake:
        fake.now.return_value = datetime(2019, 10, 31, 18, 6, 4)
        assert inst.value == 123
        assert inst.expires == datetime(2019, 10, 31, 18, 6, 5)
        assert inst.not_expired()
        assert not inst.expired()

    with patch('trivial_tools.storage.caching_instance.datetime') as fake:
        fake.now.return_value = datetime(2019, 10, 31, 18, 6, 5)
        assert inst.value == 123
        assert inst.expires == datetime(2019, 10, 31, 18, 6, 5)
        assert inst.expired()
        assert not inst.not_expired()


def test_repr():
    """
    Проверка преобразования в текст
    """
    inst = CachingInstance(123)
    assert str(inst) == 'CachingInstance(value=123, expires=None)'

    with patch('trivial_tools.storage.caching_instance.datetime') as fake:
        fake.now.return_value = datetime(2019, 10, 31, 18, 6, 5)
        inst = CachingInstance(345, 56)
        assert str(inst) == ('CachingInstance(value=345, '
                             'expires=datetime.datetime(2019, 10, 31, 18, 7, 1))')



# -*- coding: utf-8 -*-
"""

    Инструменты для работы с именами в АПК "ПАУК"

"""
# встроенные модули
import re
from contextlib import suppress as _suppress

# шаблон имени паука
from datetime import date
from typing import Optional

# шаблон для проверки spider_id
SPIDER_ID_PATTERN = re.compile(r'^\d{2}-\d{5}$')

# шаблон для извлечения даты из имени CSV файла
CSV_DATE_PATTERN = re.compile(r'\d{2}-\d{5}_(\d{4})-(\d{2})-(\d{2}).csv$')

# шаблон для извлечения spider_id из имени CSV файла
CSV_ID_PATTERN = re.compile(r'(\d{2}-\d{5})_\d{4}-\d{2}-\d{2}.csv$')


def spider_id_to_int(spider_id: str, pattern=re.compile(r'(\d{2})-(\d{5})')) -> int:
    """
    Конвертировать spider_id в числовое представление

    >>> spider_id_to_int('02-00241')
    200241

    >>> spider_id_to_int('99-99999')
    9999999
    """
    group, number = pattern.match(spider_id).groups()
    result = int(group) * 100_000 + int(number)
    return result


def spider_id_to_str(spider_iid: int) -> str:
    """
    Конвертировать spider_id из числового представления в строковое

    >>> spider_id_to_str(200241)
    '02-00241'

    >>> spider_id_to_str(9999999)
    '99-99999'
    """
    group, number = divmod(spider_iid, 100_000)
    result = f'{group:02d}-{number:05d}'
    return result


def spider_sorter(key: str) -> int:
    """
    Ключевая функция для сортировки пауков. Сортирует по номеру паука без учёта серии

    >>> spider_sorter('02-00241')
    241
    """
    result = -1
    if isinstance(key, str) and len(key) == 8:
        with _suppress(ValueError):
            result = int(key[3:])
    return result


def is_spider_id(string: str, pattern=SPIDER_ID_PATTERN) -> bool:
    """
    Проверить, является ли строка валидным id паука

    >>> is_spider_id('02-00125')
    True
    >>> is_spider_id('asd kjk')
    False
    """
    result = pattern.match(string) is not None
    return result


def date_from_spider_csv(filename: str, pattern=CSV_DATE_PATTERN) -> Optional[date]:
    """
    Извлечь дату из имени CSV файла паука

    >>> date_from_spider_csv('02-00183_2019-07-24.csv')
    datetime.date(2019, 7, 24)
    """
    data = pattern.search(filename)
    if data:
        year, month, day = data.groups()
        spider_date = date(int(year), int(month), int(day))
        return spider_date
    return None


def spider_id_from_spider_csv(filename: str, pattern=CSV_ID_PATTERN) -> Optional[str]:
    """
    Извлечь id паука из имени файла

    >>> spider_id_from_spider_csv('02-00183_2019-07-24.csv')
    '02-00183'

    >>> spider_id_from_spider_csv('02-00183_201s9-07-24.csv')

    """
    spider_id = pattern.search(filename)
    if spider_id:
        spider_id = spider_id.group(1)
        return spider_id
    return None


def has_csv_date(string: str, pattern=CSV_DATE_PATTERN) -> bool:
    """
    Проверить, похоже ли это на CSV файл паука с датой
    """
    result = pattern.match(string)
    return bool(result)



# -*- coding: utf-8 -*-
"""

    Инструменты обработки текстового представления дат

"""
# встроенные модули
import time
from datetime import date, datetime
from typing import Sequence, List, Union

__all__ = [
    'parse_date',
    'parse_dates',
    'parse_time',
    'parse_time_s',
    'parse_time_ms',
    'cur_time',
    'cur_time_s',
    'cur_time_ms',
    'timestamp_to_str',
    'date_to_text',
    'datetime_to_text',
    'datetime_to_text_s',
    'datetime_to_text_ms'
]


def parse_date(string: str, format_string: str = '%Y-%m-%d') -> date:
    """
    Обработка даты (при возможности)

    >>> parse_date('2019-12-01')
    datetime.date(2019, 12, 1)
    """
    return datetime.strptime(string, format_string).date()


def parse_dates(container: Sequence, format_string: str = '%Y-%m-%d') -> List[date]:
    """
    Распарсить множество дат

    >>> parse_dates(['2019-12-01', '2019-12-02'])
    [datetime.date(2019, 12, 1), datetime.date(2019, 12, 2)]
    """
    result = [parse_date(each, format_string) for each in container]
    result.sort()
    return result


def parse_time(text: str, format_string: str) -> datetime:
    """
    Преобразовать текст в объект времени
    """
    return datetime.strptime(text, format_string)


def parse_time_s(text: str, format_string: str = "%Y-%m-%d %H:%M:%S") -> datetime:
    """
    Преобразовать текст в объект времени с точностью до секунд

    >>> parse_time_s('2019-12-01 14:32:27')
    datetime.datetime(2019, 12, 1, 14, 32, 27)
    """
    return datetime.strptime(text, format_string)


def parse_time_ms(text: str, format_string: str = "%Y-%m-%d %H:%M:%S.%f") -> datetime:
    """
    Преобразовать текст в объект времени с точностью до микросекунд

    >>> parse_time_ms('2019-12-01 14:32:27.235486')
    datetime.datetime(2019, 12, 1, 14, 32, 27, 235486)
    """
    return parse_time(text, format_string)


def cur_time(format_string: str) -> str:
    """
    Получить строку с текущим временем
    """
    return datetime.now().strftime(format_string)


def cur_time_s(format_string: str = '%H:%M:%S') -> str:
    """
    Получить строку с текущим временем с точностью до секунд
    """
    return cur_time(format_string)


def cur_time_ms(format_string: str = '%H:%M:%S.%f') -> str:
    """
    Получить строку с текущим временем с точностью до микросекунд
    """
    return cur_time(format_string)


def timestamp_to_str(timestamp: float, format_string: str = "%Y-%m-%d %H:%M:%S") -> str:
    """
    Преобразовать timestamp в текстовую форму
    """
    return time.strftime(format_string, datetime.fromtimestamp(timestamp).timetuple())


def date_to_text(moment: date, format_string: str = "%Y-%m-%d") -> str:
    """
    Преобразовать время в виде date в текстовую форму
    """
    return datetime_to_text(moment, format_string)


def datetime_to_text(moment: Union[date, datetime], format_string: str) -> str:
    """
    Преобразовать время в виде datetime в текстовую форму
    """
    return moment.strftime(format_string)


def datetime_to_text_s(moment: datetime, format_string: str = "%Y-%m-%d %H:%M:%S") -> str:
    """
    Преобразовать время в виде datetime в текстовую форму
    """
    return datetime_to_text(moment, format_string)


def datetime_to_text_ms(moment: datetime, format_string: str = "%Y-%m-%d %H:%M:%S.%f") -> str:
    """
    Преобразовать время в виде datetime в текстовую форму
    """
    return datetime_to_text(moment, format_string)



# -*- coding: utf-8 -*-
"""

    Тесты ресемплировщика

"""
# встроенные модули
from datetime import datetime

# сторонние модули
import pytest

# модули проекта
from trivial_tools.containers.class_resampler import Resampler


@pytest.fixture()
def var_a():
    return datetime(2019, 12, 1, 12, 0, 0), 'A'


@pytest.fixture()
def var_b():
    return datetime(2019, 12, 1, 12, 0, 5), 'B'


@pytest.fixture()
def var_c():
    return datetime(2019, 12, 1, 12, 0, 10), 'C'


@pytest.fixture()
def var_d():
    return datetime(2019, 12, 1, 12, 0, 15), 'D'


@pytest.fixture()
def var_e():
    return datetime(2019, 12, 1, 12, 0, 20), 'E'


@pytest.fixture()
def var_items(var_a, var_b, var_c, var_d, var_e):
    return [var_a, var_b, var_c, var_d, var_e]


def test_creation():
    """
    Проверка создания
    """
    r = Resampler()
    assert r.max_gap == 30
    assert r.datetime_index == 0


def test_str(var_a):
    """
    Проверка текстового представления
    """
    r = Resampler()
    assert str(r) == 'Resampler[None]'

    r = Resampler(40, 0)
    list(r.iter_push(var_a))
    assert str(r) == 'Resampler[2019-12-01 12:00:00]'


def test_repr():
    """
    Проверка текстового представления
    """
    r = Resampler()
    assert repr(r) == 'Resampler(max_gap=30, datetime_index=0, placeholder=None)'

    r = Resampler(40, 4)
    assert repr(r) == 'Resampler(max_gap=40, datetime_index=4, placeholder=None)'

    r = Resampler(40, 4, 'Fail!')
    assert repr(r) == "Resampler(max_gap=40, datetime_index=4, placeholder='Fail!')"


def test_iter_push(var_items):
    """
    Должен выдавать генератор с набором заполнителей
    """
    r = Resampler()

    ref = 0
    for payload in var_items:
        for each in r.iter_push(payload):
            assert each[0].second == ref
            ref += 1


def test_same_time(var_a, var_b, var_c, var_d, var_e):
    """
    Проверка получения нескольких значений с одной меткой времени
    """
    r = Resampler()
    list(r.iter_push((datetime(2019, 12, 1, 12, 0, 0), 'A')))
    list(r.iter_push((datetime(2019, 12, 1, 12, 0, 0), 'B')))
    list(r.iter_push((datetime(2019, 12, 1, 12, 0, 0), 'C')))
    list(r.iter_push((datetime(2019, 12, 1, 12, 0, 0), 'D')))
    assert r.previous == [datetime(2019, 12, 1, 12, 0, 0), 'D']


def test_wrong_time():
    """
    Проверка получения значения с неправильным временем
    """
    r = Resampler()

    list(r.iter_push((datetime(2019, 12, 1, 12, 0, 9), 'A')))
    list(r.iter_push((datetime(2019, 12, 1, 12, 0, 0), 'B')))
    list(r.iter_push((datetime(2019, 12, 1, 12, 0, 15), 'C')))

    gen = r.iter_push((datetime(2019, 12, 1, 12, 0, 17), 'D'))

    assert next(gen) == [datetime(2019, 12, 1, 12, 0, 16), 'C']
    assert next(gen) == [datetime(2019, 12, 1, 12, 0, 17), 'D']


def test_empty_payload():
    """
    Проверить отсутствие срабатывания при подаче пустого значения
    """
    r = Resampler()
    list(r.iter_push(()))
    assert r.previous is None


def test_long_gap():
    """
    Проверить заполнение при слишком больших паузах
    """
    r = Resampler(max_gap=4, placeholder='null')

    package = [
        (datetime(2019, 12, 1, 12, 0, 0), 'A'),
        (datetime(2019, 12, 1, 12, 0, 4), 'B'),
        (datetime(2019, 12, 1, 12, 0, 8), 'C'),
        (datetime(2019, 12, 1, 12, 0, 15), 'D'),
        (datetime(2019, 12, 1, 12, 0, 19), 'E')
    ]

    ref = iter([
        'A', 'A', 'A', 'A', 'B',
        'B', 'B', 'B', 'C',
        'null', 'null', 'null', 'null', 'null', 'null', 'D',
        'D', 'D', 'D', 'E'
    ])

    for payload in package:
        for each in r.iter_push(payload):
            assert next(ref) == each[-1]


def test_push(var_a, var_b):
    """
    Проверка алгоритма добавления без итераций
    """
    r = Resampler(max_gap=5, placeholder='null')
    assert r.push(var_a) == []

    item_1 = r.push(var_b)
    assert item_1 == [
        [datetime(2019, 12, 1, 12, 0, 0), 'A'],
        [datetime(2019, 12, 1, 12, 0, 1), 'A'],
        [datetime(2019, 12, 1, 12, 0, 2), 'A'],
        [datetime(2019, 12, 1, 12, 0, 3), 'A'],
        [datetime(2019, 12, 1, 12, 0, 4), 'A'],
        [datetime(2019, 12, 1, 12, 0, 5), 'B']
    ]

    r.clear()

    assert r.push(var_a) == []
    for each_x, each_y in zip(item_1, r.iter_push(var_b)):
        assert each_x == each_y


def test_clear(var_a, var_b):
    """
    Проверка на стирание памяти
    """
    ref = [
        [datetime(2019, 12, 1, 12, 0, 0), 'A'],
        [datetime(2019, 12, 1, 12, 0, 1), 'A'],
        [datetime(2019, 12, 1, 12, 0, 2), 'A'],
        [datetime(2019, 12, 1, 12, 0, 3), 'A'],
        [datetime(2019, 12, 1, 12, 0, 4), 'A'],
        [datetime(2019, 12, 1, 12, 0, 5), 'B']
    ]
    r = Resampler(max_gap=5, placeholder='null')

    assert r.push(var_a) == []
    assert r.push(var_b) == ref

    r.clear()

    assert r.push(var_a) == []
    assert r.push(var_b) == ref

    r.clear()

    assert r.push(var_a) == []
    assert r.push(var_b) == ref



# -*- coding: utf-8 -*-
"""

    Класс для ресемплирования данных с неоднородными метками времени в посекундный формат

    Пример работы с контейнером:

    r = Resampler()
    result = r.push((datetime(2019, 12, 1, 12, 0, 0), 'A'))  # result будет []
    result = r.push((datetime(2019, 12, 1, 12, 0, 4), 'B'))

    В result будет лежать:
    [
        [datetime(2019, 12, 1, 12, 0, 0), 'A'],
        [datetime(2019, 12, 1, 12, 0, 1), 'A'],
        [datetime(2019, 12, 1, 12, 0, 2), 'A'],
        [datetime(2019, 12, 1, 12, 0, 3), 'A'],
        [datetime(2019, 12, 1, 12, 0, 4), 'B']
    ]

"""
# встроенные модули
from datetime import timedelta
from typing import Generator, Any

# модули проекта
from trivial_tools.formatters.base import s_type
from trivial_tools.datetime_tools.text import datetime_to_text_s


class Resampler:
    """
    Выдаёт данные в посекундном виде.
    Создаётся по одному экземпляру на источник данных
    Сами данные не хранит, работает по двум меткам времени
    """
    __slots__ = ('max_gap', 'datetime_index', 'previous', 'placeholder', '_full_on')

    def __init__(self, max_gap: int = 30, datetime_index: int = 0, placeholder: Any = None):
        """
        Создание экземпляра

        :param max_gap: предельно допустимое количество секунд между временными метками. Периоды
        больше этого времени считаются периодом сбоя связи и забиваются nan

        :param datetime_index: объект ожидает получать некоторые данные в виде кортежа,
        соответственно datetime_index это индекс в этом кортеже,
        по которому можно найти метку времени.
        :param placeholder: заполнитель для случая, когда у нас не нашлось значения для подстановки
        :param _full_on: флаг, показывающий, что ресемплер вошёл в рабочий режим. Без него
        будет съедено первое значение в списке
        """
        self.max_gap = max_gap
        self.datetime_index = datetime_index
        self.placeholder = placeholder
        self.previous = None
        self._full_on = False

    def __repr__(self) -> str:
        """
        Текстовое представление
        """
        result = (
            f'{s_type(self)}('
            f'max_gap={self.max_gap}, '
            f'datetime_index={self.datetime_index}, '
            f'placeholder={self.placeholder!r})'
        )
        return result

    def __str__(self) -> str:
        """
        Текстовое представление
        """
        if self.previous:
            previous = datetime_to_text_s(self.previous[self.datetime_index])
        else:
            previous = 'None'

        result = f'{s_type(self)}[{previous}]'
        return result

    def iter_push(self, payload: tuple) -> Generator[tuple, None, None]:
        """
        Добавить новый кусочек данных с меткой времени и получить обратно итератор по
        ресемплированным данным

        Подразумевается, что данные уже округлены до целых секунд

        :param payload: кортеж данных с временной меткой
        :return: последовательность кортежей с ресемплированными данными
        """
        # ----------------- Первичная инициализация, мы пока не можем считать
        if not payload:
            return None

        payload = list(payload)

        # у нас нет старого значения, берём это без проверки
        if self.previous is None:
            self.previous = payload
            return None

        new_seconds = self.get_seconds(payload)
        old_seconds = self.get_seconds(self.previous)

        if new_seconds == old_seconds:
            # при равенстве мы всё-равно замещаем старые показания новыми
            self.previous = payload
            return None

        if new_seconds < old_seconds:
            # это значение слишком старое
            return None

        old_moment = self.previous[self.datetime_index]
        delta = new_seconds - old_seconds

        if delta > self.max_gap:
            # забиваем пустыми значениями т.к. превышен предельный интервал
            body = [self.placeholder if i != self.datetime_index else x
                    for i, x in enumerate(self.previous)]
        else:
            # вытаскиваем значения (идут после индекса времени)
            body = list(self.previous)

        # первое значение в списке требует особого отношения
        if not self._full_on:
            yield self.previous
            self._full_on = True

        # генерируем серию промежуточных значений
        for i in range(1, delta):
            moment = old_moment + timedelta(seconds=i)
            body[self.datetime_index] = moment
            yield list(body)

        yield payload

        self.previous = payload

    def push(self, payload: tuple) -> list:
        """
        Аналогично итеративному варианту, только получить сразу список

        :param payload: кортеж данных с временной меткой
        :return: список кортежей с ресемплированными данными
        """
        result = []
        for each in self.iter_push(payload):
            result.append(each)
        return result

    def get_seconds(self, payload: list) -> int:
        """
        Выделить число секунд для datetime
        """
        return int(payload[self.datetime_index].timestamp())

    def clear(self):
        """
        Сбросить в исходное состояние
        """
        self.previous = None
        self._full_on = False



# -*- coding: utf-8 -*-
"""

    Тесты карусели

"""
# сторонние модули
import pytest

# модули проекта
from trivial_tools.containers.class_carousel import Carousel


def test_creation():
    """
    Проверка создания
    """
    with pytest.raises(ValueError):
        Carousel()

    c = Carousel(window=2)
    assert len(c) == 0
    assert c.window == 2
    assert c.get_contents() == []

    c = Carousel([1, 2, 3])
    assert len(c) == 3
    assert c.window == 3
    assert c.get_contents() == [1, 2, 3]
    assert c.extract() == [1, 2, 3]

    c = Carousel([1, 2, 3], window=2)
    assert len(c) == 2
    assert c.window == 2
    assert c.get_contents() == [2, 3]
    assert c.extract() == [2, 3]


def test_push():
    """
    Проверка добавления элемента
    """
    maxlen = 3
    c = Carousel(window=maxlen, sentinel=None)
    assert c.get_contents() == []

    x = c.push(1)
    assert c.get_contents() == [1]
    assert x is None

    x = c.push(2)
    assert c.get_contents() == [1, 2]
    assert x is None

    x = c.push(3)
    assert c.get_contents() == [1, 2, 3]
    assert x is None

    x = c.push(4)
    assert c.get_contents() == [2, 3, 4]
    assert x == 1

    x = c.push(5)
    assert c.get_contents() == [3, 4, 5]
    assert x == 2

    x = c.push(6)
    assert c.get_contents() == [4, 5, 6]
    assert x == 3


def test_len():
    """
    Проверка измерения длины
    """
    maxlen = 5
    c = Carousel(window=maxlen)
    for i in range(1, 10):
        c.push(i)
        assert len(c) == i if i < maxlen else maxlen
    assert len(c) == maxlen


def test_get_contents():
    """
    Не должно влиять на внутренности
    """
    c = Carousel([1, 2, 3])
    assert c.get_contents() == [1, 2, 3]
    assert c._data == [1, 2, 3]
    assert c._data == c.get_contents()

    c = Carousel([1, 2, 3], window=2)
    assert c.get_contents() == [2, 3]
    assert c._data == [3, 2]
    assert c._data != c.get_contents()


def test_str():
    """
    Проверка текстового представления
    """
    c = Carousel(window=4)
    assert str(c) == 'Carousel([], window=4)'

    c = Carousel([1, 2, 3])
    assert str(c) == 'Carousel([1, 2, 3], window=3)'

    c = Carousel([1, 2, 3], window=2)
    assert str(c) == 'Carousel([2, 3], window=2)'


def test_str_long():
    """
    Проверка текстового представления
    """
    c = Carousel(window=40)
    assert str(c) == 'Carousel([], window=40)'

    c = Carousel(range(100), window=40)
    assert str(c) == 'Carousel([60, 61, 62, ..., 97, 98, 99], len=40, window=40)'


def test_repr_long():
    """
    Проверка текстового представления
    """
    c = Carousel(window=40)
    assert repr(c) == 'Carousel([NULL, NULL, NULL, ..., NULL, NULL, NULL], len=0, window=40)'

    c = Carousel(range(100), window=40)
    assert repr(c) == 'Carousel([60, 61, 62, ..., 97, 98, 99], len=40, window=40)'


def test_resize():
    """
    Проверка изменения размера
    """
    c = Carousel([1, 2, 3], window=2)
    assert len(c) == 2
    assert c.window == 2
    assert c.get_contents() == [2, 3]

    assert repr(c) == 'Carousel([2, 3], window=2)'
    c.resize(5)
    assert repr(c) == 'Carousel([2, 3, NULL, NULL, NULL], window=5)'

    c.push(4)
    c.push(5)

    assert len(c) == 4
    assert c.window == 5
    assert c.get_contents() == [2, 3, 4, 5]
    assert repr(c) == 'Carousel([2, 3, 4, 5, NULL], window=5)'

    c.push(6)

    assert len(c) == 5
    assert c.window == 5
    assert c.get_contents() == [2, 3, 4, 5, 6]
    assert repr(c) == 'Carousel([2, 3, 4, 5, 6], window=5)'

    c.resize(2)
    assert repr(c) == 'Carousel([5, 6], window=2)'

    assert len(c) == 2
    assert c.window == 2
    assert c.get_contents() == [5, 6]


def test_resize_empty():
    """
    Проверка изменения размера пустой карусели
    """
    c = Carousel(window=2)
    assert c.window == 2
    assert repr(c) == 'Carousel([NULL, NULL], window=2)'

    c.resize(5)
    assert c.window == 5
    assert repr(c) == 'Carousel([NULL, NULL, NULL, NULL, NULL], window=5)'

    c.resize(3)
    assert c.window == 3
    assert repr(c) == 'Carousel([NULL, NULL, NULL], window=3)'


def test_no_resize():
    """
    Проверка изменения размера при том же значении
    """
    c = Carousel(window=2)
    assert c.window == 2
    c.resize(c.window)
    assert c.window == 2


def test_iter():
    """
    Проверка итерирования по карусели
    """
    base = [1, 2, 3]
    c = Carousel(base, window=4)
    for ref, value in zip(base, c):
        assert ref == value

    # убедимся, что не сломали список
    assert base == [1, 2, 3]


def test_getitem_int():
    """
    Проверка обращения к элементу (целое как индекс)
    """
    c = Carousel(window=4)

    with pytest.raises(IndexError):
        assert c[0]

    with pytest.raises(IndexError):
        # noinspection PyTypeChecker
        assert c['test']

    c.push(1)
    assert c[0] == 1
    assert c[-1] == 1

    with pytest.raises(IndexError):
        assert c[2]

    c.push(2)
    assert c[0] == 1
    assert c[1] == 2
    assert c[-1] == 2
    assert c[-2] == 1

    c.push(3)
    assert c[0] == 1
    assert c[1] == 2
    assert c[2] == 3
    assert c[-1] == 3
    assert c[-2] == 2
    assert c[-3] == 1

    c.push(4)
    assert c[0] == 1
    assert c[1] == 2
    assert c[2] == 3
    assert c[3] == 4
    assert c[-1] == 4
    assert c[-2] == 3
    assert c[-3] == 2
    assert c[-4] == 1

    c.push(5)
    assert c[0] == 2
    assert c[1] == 3
    assert c[2] == 4
    assert c[3] == 5
    assert c[-1] == 5
    assert c[-2] == 4
    assert c[-3] == 3
    assert c[-4] == 2

    c.push(6)
    assert c[0] == 3
    assert c[1] == 4
    assert c[2] == 5
    assert c[3] == 6
    assert c[-1] == 6
    assert c[-2] == 5
    assert c[-3] == 4
    assert c[-4] == 3

    c.push(7)
    assert c[0] == 4
    assert c[1] == 5
    assert c[2] == 6
    assert c[3] == 7
    assert c[-1] == 7
    assert c[-2] == 6
    assert c[-3] == 5
    assert c[-4] == 4

    c.push(8)
    assert c[0] == 5
    assert c[1] == 6
    assert c[2] == 7
    assert c[3] == 8
    assert c[-1] == 8
    assert c[-2] == 7
    assert c[-3] == 6
    assert c[-4] == 5


def test_getitem_slice():
    """
    Проверка обращения к элементу (срез как индекс)
    """
    c = Carousel(window=4)

    assert c[:] == []

    for i in range(1, 9):
        c.push(i)

    assert c[0] == 5
    assert c[1] == 6
    assert c[2] == 7
    assert c[3] == 8

    assert c[:] == [5, 6, 7, 8]

    assert c[:4] == [5, 6, 7, 8]
    assert c[:3] == [5, 6, 7]
    assert c[:2] == [5, 6]
    assert c[:1] == [5]

    assert c[1:] == [6, 7, 8]
    assert c[2:] == [7, 8]
    assert c[3:] == [8]
    assert c[4:] == []

    c.push(9)

    assert c[:] == [6, 7, 8, 9]

    assert c[:4] == [6, 7, 8, 9]
    assert c[:3] == [6, 7, 8]
    assert c[:2] == [6, 7]
    assert c[:1] == [6]

    assert c[1:] == [7, 8, 9]
    assert c[2:] == [8, 9]
    assert c[3:] == [9]
    assert c[4:] == []

    c.push(10)

    assert c[:] == [7, 8, 9, 10]

    assert c[1:] == [8, 9, 10]
    assert c[2:] == [9, 10]
    assert c[3:] == [10]
    assert c[4:] == []

    assert c[0:3:2] == [7, 9]


def test_setitem_int():
    """
    Проверка подмены элемента
    """
    c = Carousel(window=3)

    with pytest.raises(IndexError):
        c[0] = 1

    c.push(1)
    c.push(2)
    c.push(3)
    c.push(4)
    c.push(5)
    c.push(6)

    assert c[0] == 4
    assert c[1] == 5
    assert c[2] == 6

    assert c[-1] == 6
    assert c[-2] == 5
    assert c[-3] == 4

    assert c[:] == [4, 5, 6]

    c[0] = 9
    assert c[0] == 9
    assert c[1] == 5
    assert c[2] == 6

    c[1] = 7
    assert c[0] == 9
    assert c[1] == 7
    assert c[2] == 6

    c[2] = 4
    assert c[0] == 9
    assert c[1] == 7
    assert c[2] == 4

    c[-1] = 78
    assert c[0] == 9
    assert c[1] == 7
    assert c[2] == 78

    with pytest.raises(IndexError):
        c[75] = 6


def test_setitem_slice():
    """
    Проверка подмены элемента
    """
    c = Carousel(window=4)

    with pytest.raises(IndexError):
        c[:] = [9, 8, 7, 6]


def test_is_sentinel():
    """
    Проверка выдачи пустого элемента
    """
    c = Carousel(window=2)

    with pytest.raises(IndexError):
        _ = c[0]



# -*- coding: utf-8 -*-
"""

    Тесты инструментов запуска

"""
import os
import sys


def test_test():
    print()
    print('sys.argv', sys.argv)
    print('os.getcwd()', os.getcwd())
    print('sys.executable', sys.executable)




# -*- coding: utf-8 -*-
"""

    Инструменты запросов

"""
# встроенные модули
from typing import Dict, Any, Optional

# сторонние модули
import requests
from loguru import logger

# модули проекта
from trivial_tools.runners.base import repeat_on_exceptions


def make_post(url: str, json: dict) -> Optional[Dict[str, Any]]:
    """
    Выполнить POST запрос и получить данные
    """
    response = requests.post(url, json=json)

    if response.status_code == 200:
        result = response.json()
    else:
        logger.critical(f'Код возврата: {response.status_code}')
        logger.critical(response.content.decode('utf-8'))
        result = None

    return result


@repeat_on_exceptions(repeats=0, case=requests.ConnectionError, delay=1.0)
def make_safe_post(url: str, json: dict) -> Optional[Dict[str, Any]]:
    """
    Безопасный вариант POST запроса
    """
    return make_post(url, json)



# -*- coding: utf-8 -*-
"""

    Инструменты простых вычислений

"""
# встроенные модули
import math
from typing import Union

SUFFIXES = {
    'RU': {'B': 'Б', 'kB': 'кБ', 'MB': 'МБ', 'GB': 'ГБ', 'TB': 'ТБ', 'PB': 'ПБ', 'EB': 'ЭБ',
           'KiB': 'КиБ', 'MiB': 'МиБ', 'GiB': 'ГиБ', 'TiB': 'ТиБ', 'PiB': 'ПиБ', 'EiB': 'ЭиБ'},
    'EN': {'B': 'B', 'kB': 'kB', 'MB': 'MB', 'GB': 'GB', 'TB': 'TB', 'PB': 'PB', 'EB': 'EB',
           'KiB': 'KiB', 'MiB': 'MiB', 'GiB': 'GiB', 'TiB': 'TiB', 'PiB': 'PiB', 'EiB': 'EiB'},
}


def math_round(number: float, decimals: int = 0) -> float:
    """Округлить математическиим (не банковским) способом.

    Работает обычным математическим образом, в отличие от встроенной функции round(),
    которая использует банковское округление.

    :param number: число, которое требуется округлить
    :param decimals: сколько разрядов после запятой оставить
    :return: округлённое число с плавающей запятой

    >>> math_round(2.735, 2)
    2.74
    >>> round(2.735, 2)
    2.73
    """
    if math.isnan(number):
        return math.nan

    exp = number * 10 ** decimals
    if abs(exp) - abs(math.floor(exp)) < 0.5:
        return math.floor(exp) / 10 ** decimals
    return math.ceil(exp) / 10 ** decimals


def sep_digits(number: Union[int, float, str], precision: int = 2) -> str:
    """
    Вывести число с разделением на разряды

    >>> sep_digits('12345678')
    '12345678'

    >>> sep_digits(12345678)
    '12 345 678'

    >>> sep_digits(1234.5678)
    '1 234.57'

    >>> sep_digits(1234.5678, precision=4)
    '1 234.5678'
    """
    if isinstance(number, int):
        result = '{:,}'.format(number).replace(',', ' ')

    elif isinstance(number, float):
        result = '{:,}'.format(math_round(number, precision)).replace(',', ' ')
        if '.' in result:
            tail = result.rsplit('.')[-1]
            result += '0' * (precision - len(tail))

    else:
        result = str(number)

    return result


def byte_count_to_si(total_bytes: int, lang: str = 'RU') -> str:
    """
    Перевести число байт в укороченную форму согласно системы СИ, по основанию 1000

    :param total_bytes: целое число байт
    :param lang: язык, на котором должен быть написан суффикс
    :return: текстовое представление величины

    >>> byte_count_to_si(1023)
    '1.0 kB'
    """
    prefix = ''
    if total_bytes < 0:
        prefix = '-'
        total_bytes = abs(total_bytes)

    if total_bytes < 1000:
        return f'{prefix}{int(total_bytes)} ' + SUFFIXES[lang]['B']

    if total_bytes < 999_950:
        return f'{prefix}{total_bytes / 1000:0.1f} ' + SUFFIXES[lang]['kB']

    total_bytes /= 1000

    if total_bytes < 999_950:
        return f'{prefix}{total_bytes / 1000 :0.1f} ' + SUFFIXES[lang]['MB']

    total_bytes /= 1000

    if total_bytes < 999_950:
        return f'{prefix}{total_bytes / 1000 :0.1f} ' + SUFFIXES[lang]['GB']

    total_bytes /= 1000

    if total_bytes < 999_950:
        return f'{prefix}{total_bytes / 1000 :0.1f} ' + SUFFIXES[lang]['TB']

    return f'{prefix}{total_bytes / 1_000_000_000 :0.1f} ' + SUFFIXES[lang]['EB']


def byte_count_to_text(total_bytes: int, lang: str = 'RU') -> str:
    """
    Перевести число байт в укороченную форму по основанию 1024

    :param total_bytes: целое число байт
    :param lang: язык, на котором должен быть написан суффикс
    :return: текстовое представление величины

    >>> byte_count_to_text(1023)
    '1023 B'
    """
    prefix = ''
    if total_bytes < 0:
        prefix = '-'
        total_bytes = abs(total_bytes)

    if total_bytes < 1024:
        return f'{prefix}{int(total_bytes)} ' + SUFFIXES[lang]['B']

    total_bytes /= 1024

    if total_bytes < 1024:
        return f'{prefix}{total_bytes:0.1f} ' + SUFFIXES[lang]['KiB']

    total_bytes /= 1024

    if total_bytes < 1024:
        return f'{prefix}{total_bytes:0.1f} ' + SUFFIXES[lang]['MiB']

    total_bytes /= 1024

    if total_bytes < 1024:
        return f'{prefix}{total_bytes:0.1f} ' + SUFFIXES[lang]['GiB']

    total_bytes /= 1024

    if total_bytes < 1024:
        return f'{prefix}{total_bytes:0.1f} ' + SUFFIXES[lang]['TiB']

    return f'{total_bytes / 1024 / 1024 :0.1f} ' + SUFFIXES[lang]['EiB']



# -*- coding: utf-8 -*-
"""

    Работа с переменными среды

"""
# встроенные модули
import os
import sys

# сторонние модули
from loguru import logger


def get_env_variable(key: str) -> str:
    """
    Получить переменную окружения
    """
    value = os.getenv(key)

    if value is None:
        logger.critical(f'Не установлена переменная окружения "{key}"')
        sys.exit(1)

    return value


def get_path_from_env(key: str) -> str:
    """
    Обратиться к переменным окружения и получить сохранённый путь
    """
    path = get_env_variable(key)

    if not os.path.exists(path):
        logger.critical(f'Не найден каталог по пути "{path}"')
        sys.exit(1)

    return path


def get_full_path_from_env(key: str, *args: str, auto_create: bool = True) -> str:
    """
    Получить путь к каталогу по ключу переменной среды
    При наличии суффикса мы добавляем его в путь как подкаталог
    """
    orig_path = get_path_from_env(key)
    if args:
        path = os.path.join(orig_path, *args)

        if not os.path.exists(path) and auto_create:
            new_path = orig_path
            for arg in args:

                if '.' in arg:
                    # содаём только каталоги!
                    continue

                new_path = os.path.join(new_path, arg)
                os.mkdir(new_path)
    else:
        path = orig_path

    return path



# -*- coding: utf-8 -*-
"""

    Тесты кеширующей машины

"""
# встроенные модули
from datetime import datetime
from unittest.mock import patch

# сторонние модули
import pytest

# модули проекта
from trivial_tools.storage.caching_machine import CachingMachine


@pytest.fixture()
def machine_persistent():
    """
    Машина без срока хранения
    """
    return CachingMachine()


@pytest.fixture()
def machine_not_persistent():
    """
    Машина со сроком хранения 5 секунд
    """
    return CachingMachine(expiration=5)


@pytest.fixture()
def machine_tiny():
    """
    Машина со сроком хранения 5 секунд и всего на 4 элемента
    """
    return CachingMachine(expiration=5, max_items=4)


def test_fill(machine_tiny):
    """
    Проверка заполнения машины
    """
    with patch('trivial_tools.storage.caching_instance.datetime') as fake:
        fake.now.return_value = datetime(2019, 10, 31, 18, 6, 0)
        assert machine_tiny.total() == 0

        machine_tiny.set('key_1', 'value')
        assert machine_tiny.total() == 1

        machine_tiny.set('key_1', 'value')
        assert machine_tiny.total() == 1

        machine_tiny.set('key_2', 'value')
        assert machine_tiny.total() == 2

        machine_tiny.set('key_3', 'value')
        assert machine_tiny.total() == 3

        machine_tiny.set('key_4', 'value')
        assert machine_tiny.total() == 4

        machine_tiny.set('key_5', 'value')
        assert machine_tiny.total() == 4

        machine_tiny.clear()
        assert machine_tiny.total() == 0


def test_caching(machine_persistent):
    """
    Проверка работы декоратора
    """
    @machine_persistent.cache_call()
    def func(x, y):
        return x + y

    assert func(1, 2) == 3
    assert func(2, 3) == 5
    assert func(1, 2) == 3

    assert machine_persistent.keys() == [('func', (1, 2)), ('func', (2, 3))]


def test_caching_string(machine_persistent):
    """
    Проверка работы декоратора
    """

    @machine_persistent.cache_call_using_strings()
    def func(x, y):
        return x + y

    assert func(1, 2) == 3
    assert func(2, 3) == 5
    assert func(1, 2) == 3

    assert machine_persistent.keys() == ['func_1_2', 'func_2_3']



# -*- coding: utf-8 -*-
"""

    Обработка данных для пауков

"""
# встроенные модули
import re
from typing import Tuple


CHANNEL_PATTERN = re.compile(r'(\d+)([ABC])(x(.*)?)?')


def analyze_channels(string: str) -> Tuple[int, str, float]:
    """
    Разложить строку каналов на составляющие
    """
    number, phase, _, mul = CHANNEL_PATTERN.search(string).groups()
    return int(number), phase.upper(), 1.0 if mul is None else float(mul)



# -*- coding: utf-8 -*-
"""

    Тесты класса с динамическими параметрами

"""
# сторонние модули
import pytest

from trivial_tools.classes.fluid import Fluid


def test_creation():
    """
    Создание экземпляра
    """
    f = Fluid(a=1, b=2)
    assert f.a == 1
    assert f.b == 2
    assert vars(f) == {'a': 1, 'b': 2}


def test_creation_wrong():
    """
    Создание экземпляра
    """
    with pytest.raises(ValueError):
        Fluid(1, 2)


def test_creation_dict():
    """
    Создание экземпляра
    """
    source = dict(a=1, b=2, c='3')
    f1 = Fluid(**source)
    f2 = Fluid.from_dict(source)
    assert f1 == f2
    assert f1.to_dict() == source
    assert f2.to_dict() == source


def test_alter():
    """
    Изменение
    """
    source = dict(a=1, b=2, c='3')
    f1 = Fluid(**source)
    assert f1.to_dict() == {'a': 1, 'b': 2, 'c': '3'}
    f2 = f1.alter(c='test', d=285)
    assert f2.to_dict() == {'a': 1, 'b': 2, 'c': 'test', 'd': 285}
    assert f1.to_dict() == {'a': 1, 'b': 2, 'c': '3'}


def test_eq():
    """
    Проверка равенства
    """
    f1 = Fluid(a=1, b=2)
    f2 = Fluid(a=1, b=2)
    assert f1 == f2
    assert f1 != 1
    assert f2 != dict(a=1, b=2)


def test_str():
    """
    Текстовое представление
    """
    f = Fluid(par_1=285, par_2='test', par_3=None)
    assert str(f) == 'Fluid(par_1=285, par_2=test, par_3=None)'

    f = Fluid(par_1=[1, 2, 3], par_2='test', par_3=None, par_4=dict(a=1), par_5=[1, 2, 3, 4, 5, 6])
    assert str(f) == """
Fluid(
    01. par_1 = [1, 2, 3]
    02. par_2 = test
    03. par_3 = None
    04. par_4 = {'a': 1}
    05. par_5 = [
                 1,
                 2,
                 3,
                 4,
                 5,
                 6
                ]
)
    """.strip()


def test_repr():
    """
    Текстовое представление
    """
    f = Fluid(par_1=285, par_2='test', par_3=None)
    assert repr(f) == "Fluid(par_1=285, par_2='test', par_3=None)"



# -*- coding: utf-8 -*-
"""

    Тесты конфига

"""
# модули проекта
from trivial_tools.config_handling.base_config import BaseConfig


def test_creation():
    """
    Создание экземпляра
    """
    c = BaseConfig(a=1, b=2)
    assert c.a == 1
    assert c.b == 2


def test_context():
    """
    Создание экземпляра
    """
    c = BaseConfig(a=1, b=2)
    BaseConfig.register(c)

    instance = BaseConfig.get_instance()

    assert instance is c
    assert instance.a == c.a
    assert instance.b == c.b

    with BaseConfig.temporary_config(a='test', b='value'):
        instance = BaseConfig.get_instance()
        assert instance.a == 'test'
        assert instance.b == 'value'

    instance = BaseConfig.get_instance()
    assert instance is c
    assert instance.a == c.a
    assert instance.b == c.b



# -*- coding: utf-8 -*-
"""

    Контейнер для вычисления скользящей суммы.
    В него можно постоянно можно добавлять элементы и быстро получать сумму

    Пример работы с контейнером:
    m = MovingSum(window=4)  # сумма равна 0
    m.push(8)    # сумма равна 8
    m.push(41)   # сумма равна 49
    m.push(9)    # сумма равна 58
    m.push(6)    # сумма равна 64
    m.push(32)   # сумма равна 88
    m.push(58)   # сумма равна 105
    m.push(8)    # сумма равна 104
    m.push(78)   # сумма равна 176
    m.push(3)    # сумма равна 147
    m.push(2)    # сумма равна 91

"""
# встроенные модули
from typing import Union, Optional, Sequence, Any

# модули проекта
from trivial_tools.containers.class_carousel import Carousel


class MovingSum(Carousel):
    """
    Контейнер для вычисления скользящей суммы.
    В него можно постоянно можно добавлять элементы и быстро получать сумму
    """
    __slots__ = ('_sum',)

    def __init__(self, source: Optional[Sequence] = None,
                 window: int = 0, sentinel: Any = object()):
        """
        Создание экземпляра

        :param source: исходная коллекция элементов, на базе которой надо собрать экземпляр
        :param window: максимальное киличество элементов (ширина окна вычисления)
        :param sentinel: элемент для заполнения пустых ячеек (можно добавить свой)
        """
        self._sum: Union[int, float] = 0
        super().__init__(source, window, sentinel)

    def push(self, value: Union[int, float]) -> None:
        """
        Добавить элемент в контейнер

        Мы выталкиваем предыдущий элемент и меняем сумму соответствующим образом. Причина
        для такого выполнения - избежать необходимости итерироваться по коллекции при
        попытке вычислить среднее значение

        :param value: новое число, которое соответствует сдвигу окна среднего на один элемент
        """
        old_value = super().push(value)

        if old_value is not self._sentinel:
            self._sum -= old_value

        self._sum += value

    def restore(self) -> None:
        """
        Сбросить параметры
        """
        self._sum = 0
        super().restore()

    @property
    def sum(self) -> float:
        """
        Сумма всех элементов
        """
        return self._sum

    def __setitem__(self, key: Union[int, slice], value: Union[int, float, Sequence]) -> None:
        """
        Записать элемент по индексу.
        Обеспечивается обычный доступ к внутреннему хранилищу, просто со смещением индекса

        :param key: ключ индексации
        :param value: данные для записи (любой тип)
        """
        if isinstance(key, int):
            self._sum -= self._data[self.get_real_index(key)]
            self._sum += value

        super().__setitem__(key, value)



# -*- coding: utf-8 -*-
"""

    Инструменты запуска

"""
# встроенные модули
import os
import time
from datetime import datetime
from functools import wraps
from typing import Callable, Union, Tuple, Any, Type, Optional, Dict

# сторонние модули
from loguru import logger

# модули проекта
from trivial_tools.special.special import fail
from trivial_tools.system.envs import get_full_path_from_env
from trivial_tools.config_handling.base_config import BaseConfig


def repeat_on_exceptions(repeats: int, case: Union[Exception, Tuple[Exception]],
                         delay: float = 1.0, verbose: bool = True) -> Callable:
    """Декоратор, заставляющий функцию повторить операцию при выбросе исключения.

    Инструмент добавлен для возможности повторной отправки HTTP запросов на серверах

    :param verbose: логгировать произошедшие исключения и выводить их на экран
    :param delay: время ожидания между попытками
    :param repeats: сколько раз повторить (0 - повторять бесконечно)
    :param case: при каком исключении вызывать повтор
    :return: возвращает фабрику декораторов
    """

    def decorator(func: Callable) -> Callable:
        """
        Декоратор
        """

        @wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            """
            Враппер
            """
            iteration = 0
            while True:
                try:
                    result = func(*args, **kwargs)
                    break
                except case as exc:
                    iteration += 1

                    if verbose:
                        logger.warning(f'{exc} в функции {func.__name__}'
                                       f' (итерация {iteration})')

                    if repeats and iteration > repeats:
                        fail(exc)

                    time.sleep(delay)

            return result

        return wrapper

    return decorator


def start_working(folder_name: str, config_name: Optional[str], func: Callable,
                  config_type: Type[BaseConfig],
                  config_filename: str = 'config.json', infinite: bool = True) -> None:
    """Стартовать скрипт с выбранными настройками.

    :param folder_name: имя суб-каталога со скриптом. Запуск предполагается из коневого каталога
    :param config_name: принудительный выбор конфигурации для скрипта при запуске из консоли
    :param func: рабочая функция скрпта, которая будет выполнять всю полезную работу
    :param config_type: класс конфигурации для выбранного скрипта
    :param config_filename: имя json файла в котором следует искать кофигурацию
    :param infinite: флаг бесконечного перезапуска скрипта при выбросе исключения
    """
    config = config_type.from_json(
        config_name=config_name,
        filename=os.path.join(folder_name, config_filename)
    )

    logger.add(
        sink=get_full_path_from_env(config.db_path, config.logger_filename),
        level=config.logger_level,
        rotation=config.logger_rotation
    )

    separator = '#' * 79

    logger.warning(separator)
    logger.warning(config.start_message)
    logger.info(f'Параметры скрипта:\n{config}')

    while True:
        # noinspection PyBroadException
        try:
            message = func(config)
            logger.warning(message)

            if message == 'stop':
                break

        except KeyboardInterrupt:
            logger.warning('\nОстановка по команде с клавиатуры!')
            break
        except Exception:
            logger.exception('\nКритический сбой!')

        if not infinite:
            break

    logger.warning(separator)


def capture_exception_output(title: str) -> Optional[str]:
    """
    Перехватить описание ошибки из loguru
    """
    storage = []
    sink_id = logger.add(sink=lambda msg: storage.append(msg))
    logger.exception(title)
    logger.remove(sink_id)
    return storage[0] if storage else None


def _execute(config, handler: Callable, separator: str) -> str:
    """
    Исполнить с логгированием
    """
    logger.add(
        config.LOGGER_FILENAME,
        level=config.LOGGER_LEVEL,
        rotation=config.LOGGER_ROTATION
    )
    logger.warning(separator)
    logger.info(config.START_MESSAGE)
    logger.info(config)

    # @@@@@@@@@@@@@@@@@@@@
    message = handler()
    # @@@@@@@@@@@@@@@@@@@@

    logger.info(message)
    return message


def init_daemon(config, handler: Callable, infinite: bool = False) -> None:
    """Стартовать скрипт с выбранными настройками.

    :param config: класс, атрибуты которого являются нашими настрояками
    :param handler: рабочая функция скрпта, которая будет выполнять всю полезную работу
    :param infinite: флаг бесконечного перезапуска скрипта при выбросе исключения
    """
    # избегаем циклической ссылки
    from trivial_tools.json_rpc.calls import call_api

    separator = '#' * 79

    while True:
        normal_stop = False
        time_start = datetime.now()

        payload: Dict[str, str] = {
            "service_name": config.SERVICE_NAME,
            "status": 'normal',
            "title": 'Успешное исполнение. Длительность - {} сек.'
        }

        # noinspection PyBroadException
        try:
            message = _execute(config, handler, separator)

            if message == '<stop>':
                logger.warning('Остановка по команде скрипта')
                normal_stop = True
                break

        except KeyboardInterrupt:
            logger.warning('Остановка по команде с клавиатуры')
            normal_stop = True
            break

        except Exception:
            description = capture_exception_output(f'Критический сбой {config.SERVICE_NAME}')
            payload['status'] = 'fail'
            payload['title'] = 'Остановка работы после {} сек. из-за критического сбоя'
            payload['text'] = description or 'Не удалось сохранить'

        else:
            payload['text'] = message

        finally:
            duration = int((datetime.now() - time_start).total_seconds())
            payload['title'] = payload['title'].format(duration)

            if config.in_production() and not normal_stop:
                # noinspection PyProtectedMember
                call_api(
                    url=config.API_URL,
                    method='register_report',
                    report=payload,
                    fingerprint=config._FINGERPRINT,
                    error_msg=f'Не удалось отправить отчёт о состоянии на {config.API_URL}!',
                    debug=config.in_debug()
                )

        if not infinite:
            break

    logger.warning(separator)



# -*- coding: utf-8 -*-
"""

    Тесты инструментов форматирования

"""
# модули проекта
from trivial_tools.formatters import *


# def test_decorate_normal():
#     assert decorate('test') == ''
#
#
# def test_decorate_symbol():
#     assert decorate('test', '%') == ''
#
#
# def test_decorate_line():
#     assert decorate('test', line_width=5) == ''


# def test_decorate_all():
#     assert decorate('test', '$', 4) == ''



# -*- coding: utf-8 -*-
"""

    Простой конфиг

"""
# встроенные модули
from typing import Any, Optional, List

# сторонние модули
from trivial_tools.runners.base import fail
from trivial_tools.formatters.base import s_type


class SimpleConfig:
    """
    Базовый класс для хранения настроек. Хранит все значения в пространстве имён класса,
    экземпляры не подразумевают наличие атрибутов.

    Особенности именования:
        АТРИБУТ - постоянное значение, не допускает изменения
        атрибут - не инициализированное значение, должно быть установлено сразу после запуска
    """
    def __setattr__(self, key: str, value: Any):
        """
        Объект неизменяемый
        """
        if key.islower():
            old_value = getattr(type(self), key)
            if isinstance(old_value, DeferredInitField):
                super().__setattr__(key, value)
                return

        fail(f'Экземпляры типа {s_type(self)} не позволяют менять атрибуты!')

    @classmethod
    def in_debug(cls) -> bool:
        """
        Находимся ли мы в режиме дебага?
        """
        return getattr(cls, 'DEBUG')

    @classmethod
    def in_production(cls) -> bool:
        """
        Находимся ли мы на продакшене?
        """
        return getattr(cls, 'APP_CONFIG') == 'production'

    def _get_attributes(self):
        """
        Получить перечень атрибутов объекта
        """
        for key in dir(self):
            if not key.startswith('__') and not callable(getattr(type(self), key)):
                yield key

    def _get_values(self):
        """
        Получить перечень значений атрибутов объекта
        """
        values = []
        for key in sorted(self._get_attributes()):
            attr = getattr(type(self), key)

            if isinstance(attr, DeferredInitField):
                if attr.already_set:
                    value = getattr(self, key)
                else:
                    value = '<Не инициализировано>'
            else:
                value = attr

            # прячем логины/пароли, которые могут попасть в конфиг
            if key.startswith('_'):
                value = '< hidden >'

            values.append((key, value))

        return values

    def _render_short(self, values: List[Any]) -> str:
        """
        Собрать текстовое представление для варианта с коротким списком аргументов
        """
        strings = [f'{key}={value}' for key, value in values]
        return f'{s_type(self)}(' + ', '.join(strings) + ')'

    def _render_long(self, values: List[Any]) -> str:
        """
        Собрать текстовое представление для варианта с длинным списком аргументов
        """
        longest = max([len(x[0]) for x in values])
        offset = '    '

        result = []
        for i, (key, value) in enumerate(values, start=1):
            if isinstance(value, (list, tuple)):
                if len(value) > 5:
                    # длинная последовательность
                    # 05. par_5 = [
                    #               1,
                    #               2,
                    string = ('{0:02d}. {1:' + str(longest) + '} = [').format(i, key)
                    result.append(offset + string)

                    for j, sub_element in enumerate(value, start=1):
                        line = offset + len(string) * ' ' + f'{sub_element}'
                        if j != len(value):
                            line += ','
                        result.append(line)

                    result.append(offset + (len(string) - 1) * ' ' + ']')
                else:
                    # короткая последовательность
                    # \t01. par_1 = [1, 2, 3]
                    string = ('{0:02d}. {1:' + str(longest) + '} = ').format(i, key)
                    result.append(offset + string + str(value))
            else:
                string = ('{0:02d}. {1:' + str(longest) + '} = {2}').format(i, key, value)
                result.append(offset + string)
        return s_type(self) + '(\n' + '\n'.join(result) + '\n)'

    def __str__(self) -> str:
        """
        Текстовое представление
        """
        values = self._get_values()

        if len(values) <= 3:
            return self._render_short(values)
        return self._render_long(values)

    def __repr__(self) -> str:
        """
        Этот класс не имеет аргументов для конструктора
        """
        return f'{s_type(self)}()'


class DeferredInitField:
    """
    Дескриптор для описания атрибутов, которые устанавливаются после запуска приложения
    """
    def __init__(self):
        """
        Инициация
        """
        self.name = None
        self.already_set = False

    def __set_name__(self, owner: SimpleConfig, name: str):
        """
        Запомнить имя дескриптора
        """
        self.name = name

    def __get__(self, instance: Optional[SimpleConfig], owner: SimpleConfig):
        """
        Получить значение атрибута
        """
        if instance is None:
            return self

        if not self.already_set:
            fail(f'Атрибут {s_type(instance)}.{self.name} ещё не был инициализирован!')

        return getattr(owner, self.name)

    def __set__(self, instance: SimpleConfig, value: Any) -> None:
        """
        Установить значение атрибута
        """
        if self.already_set:
            old_value = getattr(instance, self.name)
            fail(f'Атрибут {s_type(instance)}.{self.name} уже установлен в "{old_value!r}"!')

        # атрибут сохраняется в классе, а не в экземпляре
        setattr(type(instance), self.name, value)
        self.already_set = True



# -*- coding: utf-8 -*-
"""

    Экземпляр поля для кеширующего класса

"""
# встроенные модули
from typing import Any, Optional
from datetime import datetime, timedelta


class CachingInstance:
    """
    Экземпляры этого класса хранятся в кеширующей машине
    """
    __slots__ = ('value', 'expires')

    def __init__(self, value: Any, expires: Optional[int] = None):
        self.value = value

        if expires is None:
            self.expires = None
        else:
            self.expires = datetime.now() + timedelta(seconds=expires)

    def expired(self) -> bool:
        """
        Проверить, протухло ли значение
        """
        return not self.not_expired()

    def not_expired(self) -> bool:
        """
        Проверить, не протухло ли значение
        """
        if self.expires is None:
            return True
        return bool(self.expires > datetime.now())

    def __repr__(self):
        """
        Текстовое представление
        """
        return f'{type(self).__name__}(value={self.value!r}, expires={self.expires!r})'



# -*- coding: utf-8 -*-
"""

    Форматирование прошедшего времени в виде удобной строки

"""
# встроенные модули
from datetime import timedelta
from typing import Tuple, Optional

# модули проекта
from trivial_tools.datetime_tools.basic import cast_seconds_to_timedelta


def passed_time(delta: timedelta) -> str:
    """
    Трансформировать timedelta в описание времени в форме '6 дней, 4 часа, 35 минут'
    """
    delta = cast_seconds_to_timedelta(delta)

    weeks, rest_days = form_weeks(delta)
    delta = timedelta(days=rest_days, seconds=delta.seconds)

    days = form_days(delta)
    delta = timedelta(days=0, seconds=delta.seconds)

    hours, rest_seconds = form_hours(delta)
    delta = timedelta(days=0, hours=0, seconds=rest_seconds)

    minutes, rest_seconds = form_minutes(delta)
    delta = timedelta(days=0, hours=0, minutes=0, seconds=rest_seconds)

    seconds = form_seconds(delta)

    result = ', '.join([x for x in [weeks, days, hours, minutes, seconds] if x is not None])

    if not result:
        return '0 секунд'
    return result


def form_weeks(delta: timedelta) -> Tuple[Optional[str], int]:
    """
    Сформировать описание недель
    """
    days = delta.days

    if days < 7:
        return None, days

    weeks = days // 7
    rest = days - weeks * 7

    if weeks in (1, 21, 31, 41, 51, 61, 71, 81, 91):
        result = f'{weeks} неделя', rest

    elif weeks in (2, 3, 4):
        result = f'{weeks} недели', rest

    else:
        result = f'{weeks} недель', rest

    return result


def form_days(delta: timedelta) -> Optional[str]:
    """
    Сформировать описание дней
    """
    if not delta.days:
        return None

    days = delta.days

    if days in (1, 21, 31, 41, 51, 61, 71, 81, 91):
        result = f'{days} день'

    elif days in (2, 3, 4):
        result = f'{days} дня'

    else:
        result = f'{days} дней'

    return result


def form_hours(delta: timedelta) -> Tuple[Optional[str], int]:
    """
    Сформировать описание часов
    """
    hours = delta.seconds // 3600

    if not hours:
        return None, delta.seconds

    rest = delta.seconds % (hours * 3600)

    d1 = abs(hours) % 100
    d2 = d1 % 10

    if d2 in (1,):
        result = f'{hours} час', rest

    elif d2 in (2, 3, 4) and d1 != 14:
        result = f'{hours} часа', rest

    else:
        result = f'{hours} часов', rest

    return result


def form_minutes(delta: timedelta) -> Tuple[Optional[str], int]:
    """
    Сформировать описание минут
    """
    minutes = delta.seconds // 60

    if not minutes:
        return None, delta.seconds

    rest = delta.seconds % (minutes * 60)

    d1 = abs(minutes) % 100
    d2 = d1 % 10

    if d2 in (1,) and minutes not in (11, 12, 13):
        result = f'{minutes} минута', rest

    elif d2 in (2, 3, 4) and minutes not in (11, 12, 13, 14):
        result = f'{minutes} минуты', rest

    else:
        result = f'{minutes} минут', rest

    return result


def form_seconds(delta: timedelta) -> Optional[str]:
    """
    Сформировать описание секунд
    """
    seconds = delta.seconds

    if not seconds:
        return None

    d1 = abs(seconds) % 100
    d2 = d1 % 10

    if d2 in (1,) and seconds not in (11, 12, 13):
        result = f'{seconds} секунда'

    elif d2 in (2, 3, 4) and seconds not in (11, 12, 13):
        result = f'{seconds} секунды'

    else:
        result = f'{seconds} секунд'

    return result



# -*- coding: utf-8 -*-
"""

    Контейнер с бегущим индексом, в который постоянно можно добавлять элементы

    На базе этого контейнера предполагается собирать структуры для вычисления динамических
    характеристик непрерывного потока входных данных в реальном времени.
    Например вычисление скользящего среднего.

    Пример работы с контейнером:
    c = Carousel(window=3)  # внутреннее хранилище: [NULL, NULL, NULL]
    c.push(1)               # внутреннее хранилище: [   1, NULL, NULL]
    c.push(2)               # внутреннее хранилище: [   1,    2, NULL]
    c.push(3)               # внутреннее хранилище: [   1,    2,    3]
    c.push(4)               # внутреннее хранилище: [   4,    2,    3]
    c.push(5)               # внутреннее хранилище: [   4,    5,    3]
    c.push(6)               # внутреннее хранилище: [   4,    5,    6]
    c.push(7)               # внутреннее хранилище: [   7,    5,    6]
    c.push(8)               # внутреннее хранилище: [   7,    8,    6]
    c.get_contents() --> [6, 7, 8]

    Почему не deque или другое подобие связного списка - очень медленно.
    Предполагается, что данная реализация будет работать намного быстрее
    и меньше загружать процессор при работе.

"""
# встроенные модули
from typing import Any, Optional, Collection, List, Generator, Union

# модули проекта
from trivial_tools.special.special import fail
from trivial_tools.formatters.base import s_type


class Carousel:
    """
    Контейнер с бегущим индексом на базе списка.
    Нужен для обработки непрерывного входного потока данных
    """
    __slots__ = ('_sentinel', '_data', '_len', '_index', 'window')

    def __init__(self, source: Optional[Collection] = None,
                 window: int = 0, sentinel: Any = object()):
        """
        Создание экземпляра. Можно задать базовую коллекцию, размер и объект для пустых слотов

        :param source: исходная коллекция элементов, на базе которой надо собрать экземпляр
        :param window: максимальное количество элементов (ширина окна вычисления)
        :param sentinel: элемент для заполнения пустых ячеек (можно добавить свой)
        """
        self._sentinel = sentinel
        self._data = []
        self._len = 0
        self._index = 0

        if source:
            # есть коллекция-прототип из которой можно взять данные
            if window:
                self.window = window
            else:
                self.window = len(source)
            self.restore()
            self.populate(source)

        elif window:
            # создаём пустую карусель
            self.window = window
            self.restore()

        else:
            fail(f'Для создания экземпляра {s_type(self)} необходимо либо указать длину и/или \n'
                 'дать коллекцию из которой будет собран экземпляр.', reason=ValueError)

    def __str__(self) -> str:
        """
        Текстовый вид
        """
        return self.make_str(with_null=False)

    def __repr__(self) -> str:
        """
        Текстовый вид
        """
        return self.make_str(with_null=True)

    def make_str(self, with_null: bool = False, sep: str = ', ') -> str:
        """
        Собрать текстовое представление
        """
        contents = []

        for element in self._internals():
            if element is self._sentinel:
                if with_null:
                    contents.append('NULL')
            else:
                contents.append(str(element))

        if len(contents) <= 5:
            string = sep.join(contents)
            result = f'{s_type(self)}([{string}], window={self.window})'

        else:
            string = sep.join([*contents[0:3], '...', *contents[-3:]])
            result = f'{s_type(self)}([{string}], len={len(self)}, window={self.window})'

        return result

    def populate(self, source: Collection[Any]) -> None:
        """
        Наполнить хранилище предоставленными данными

        :param source: некоторая коллекция, элементы которой надо последовательно переложить
        """
        self._index = 0
        for value in source:
            self.push(value)

    def __len__(self) -> int:
        """
        Длина показывает сколько ячеек у нас заполнено
        """
        return self._len

    def _step_right(self, index: int) -> int:
        """
        Увеличить индекс на единицу и вернуть его значение
        """
        if index < self.window - 1:
            index += 1
        else:
            index = 0
        return index

    def push(self, element: Any) -> Any:
        """
        Добавить элемент в карусель

        Мы замещаем элемент, выталкивая его из списка и возвращаем его. Его значение потом может
        быть использовано на вызывающей стороне

        :param element: Новый элемент любого типа
        :return: Старый элемент, который был вытолкнут при добавлении (может оказаться sentinel!)
        """
        old_value = self._data[self._index]
        self._data[self._index] = element
        self._index = self._step_right(self._index)

        self._len += 1
        if self._len > self.window:
            self._len = self.window

        return old_value

    def resize(self, new_window: int) -> None:
        """
        Изменить размер карусели

        :param new_window: новая предельная длина для внутреннего хранилища
        """
        if new_window == self.window:
            return

        old_data = self.get_contents()
        self.window = new_window
        self.restore()
        self.populate(old_data)

    def get_contents(self) -> List[Any]:
        """
        Получить копию внутреннего хранилища
        """
        result = [x for x in self._internals() if x is not self._sentinel]
        return result

    def extract(self) -> List[Any]:
        """
        Получить копию внутреннего хранилища и очистить его
        """
        result = self.get_contents()
        self.restore()
        return result

    def restore(self) -> None:
        """
        Сбросить индекс и длину, заполнить
        внутреннее хранилище пустыми объектами
        """
        self._len = 0
        self._index = 0
        self._data = [self._sentinel for _ in range(self.window)]

    def _internals(self) -> Generator[Any, None, None]:
        """
        Проитерироваться по элементам внутреннего хранилища (включая пустые элементы!)
        """
        if self._len == self.window:
            yield from self._data[self._index:]
            yield from self._data[:self._index]
        else:
            yield from self._data

    def __iter__(self) -> Any:
        """
        Проитерироваться по элементам внутреннего хранилища (без пустых элементов)
        Сохраняет порядок вставки элементов
        """
        for element in self._internals():
            if element is not self._sentinel:
                yield element

    def __getitem__(self, item: Union[int, slice]) -> Any:
        """
        Обратиться к элементу по индексу.
        Обеспечивается обычный доступ к внутреннему хранилищу, просто со смещением индекса

        :param item: ключ индексации
        :return: содержимое внутреннего хранилища
        """
        if isinstance(item, int):
            if not self._len:
                fail(f'Попытка получить элемент из пустого объекта {s_type(self)}',
                     reason=IndexError)

            result = self._data[self.get_real_index(item)]

            if result is self._sentinel:
                fail(f'В экземпляре {s_type(self)} нет элемента с индексом {item!r}',
                     reason=IndexError)

            return result

        if isinstance(item, slice):
            return self.get_contents()[item]

        fail(f"Тип {s_type(self)} поддерживает работу только с индексами int и slice!",
             reason=IndexError)

    def __setitem__(self, key: Union[int, slice], value: Any) -> None:
        """
        Записать элемент по индексу.
        Обеспечивается обычный доступ к внутреннему хранилищу, просто со смещением индекса

        :param key: ключ индексации, только int
        :param value: данные для записи (любой тип)
        """
        if not isinstance(key, int):
            fail(f"Тип {s_type(self)} поддерживает работу только с индексами типа int!",
                 reason=IndexError)

        if not self._len:
            fail(f'Попытка присвоить элемент по индексу {key} в пустой объект {s_type(self)}',
                 reason=IndexError)

        self._data[self.get_real_index(key)] = value
        return

    def get_real_index(self, key: int) -> int:
        """
        Настояшее положение ячеек, с учётом их сдвига
        """
        index = self._index + key
        if index > self._len - 1:
            index -= self._len
        return index



# -*- coding: utf-8 -*-
"""

    Класс, хранящий экземпляр себя. Синглтон, для тех, кто пишет тесты

"""
# встроенные модули
from typing import TypeVar, Generic, Optional

# модули проекта
from trivial_tools.special.special import fail
from trivial_tools.formatters.base import s_type

T = TypeVar('T')


class Conservator(Generic[T]):
    """
    Класс, хранящий экземпляр себя
    """
    __instance: Optional[T] = None

    @classmethod
    def register(cls, instance: T) -> None:
        """
        Добавить валидный экземпляр
        """
        if not isinstance(instance, cls):
            fail(f'Класс {s_type(cls)} не добавляет посторонние типы!\n'
                 'При необходимости используйте метод register_force',
                 reason=ValueError)

        cls.register_force(instance)

    @classmethod
    def register_force(cls, instance: T) -> None:
        """
        Добавить валидный экземпляр
        """
        if cls.__instance:
            fail(f'Класс {s_type(cls)} уже имеет хранимый экземпляр!', reason=ValueError)

        cls.__instance = instance

    @classmethod
    def unregister(cls) -> None:
        """
        Удалить валидный экземпляр
        """
        if not cls.has_instance():
            fail(f'Класс {s_type(cls)} ещё не имеет хранимого экземпляра!', reason=ValueError)

        cls.__instance = None

    @classmethod
    def replace(cls, instance: T) -> None:
        """
        Заменить валидный экземпляр
        """
        previous = cls.__instance
        cls.unregister()
        try:
            cls.register(instance)
        except ValueError:
            cls.__instance = previous
            raise

    @classmethod
    def get_instance(cls) -> T:
        """
        Получить валидный экземпляр
        """
        if not cls.__instance:
            fail(f'Класс {s_type(cls)} ещё не имеет хранимого экземпляра!', reason=ValueError)

        return cls.__instance

    @classmethod
    def has_instance(cls) -> bool:
        """
        Проверить наличие экземпляра
        """
        return cls.__instance is not None



# -*- coding: utf-8 -*-
"""

    Тесты преобразования дат в текст и обратно

"""
# встроенные модули
from unittest.mock import patch
from datetime import datetime, date

# сторонние модули
import pytest

# модули проекта
from trivial_tools.datetime_tools.text import *


@pytest.fixture()
def timestamp():
    return 1576052808.2121792


@pytest.fixture()
def str_date1():
    return '2019-12-01'


@pytest.fixture()
def str_date2():
    return '2019-12-02'


@pytest.fixture()
def str_time_s():
    return '13:27:34'


@pytest.fixture()
def str_time_ms():
    return '13:27:34.123456'


@pytest.fixture()
def date1():
    return date(2019, 12, 1)


@pytest.fixture()
def date2():
    return date(2019, 12, 2)


@pytest.fixture()
def datetime0():
    return datetime(2019, 12, 1, 13, 27, 34)


@pytest.fixture()
def datetime1():
    return datetime(2019, 12, 1, 13, 27, 34, 123456)


@pytest.fixture()
def datetime1_str_s():
    return '2019-12-01 13:27:34'


@pytest.fixture()
def datetime1_str_ms():
    return '2019-12-01 13:27:34.123456'


def test_parse_date(date1, str_date1):
    """
    Обработка даты
    """
    assert parse_date(str_date1) == date1


def test_parse_dates(date1, date2, str_date1, str_date2):
    """
    Распарсить множество дат
    """
    assert parse_dates([str_date1, str_date2]) == [date1, date2]


def test_parse_time(datetime1, datetime1_str_s, datetime1_str_ms):
    """
    Преобразовать текст в объект времени
    """
    fmt = "%Y-%m-%d %H:%M:%S"
    assert parse_time(datetime1_str_s, fmt) == datetime.strptime(datetime1_str_s, fmt)

    fmt = "%Y-%m-%d %H:%M:%S.%f"
    assert parse_time(datetime1_str_ms, fmt) == datetime.strptime(datetime1_str_ms, fmt)


def test_parse_time_s(datetime0, datetime1_str_s):
    """
    Преобразовать текст в объект времени с точностью до секунд
    """
    assert parse_time_s(datetime1_str_s) == datetime0


def test_parse_time_ms(datetime1, datetime1_str_ms):
    """
    Преобразовать текст в объект времени с точностью до микросекунд
    """
    assert parse_time_ms(datetime1_str_ms) == datetime1


@patch('trivial_tools.datetime_tools.text.datetime')
def test_cur_time(mocked_time, datetime1, str_time_s):
    """
    Получить строку с текущим временем
    """
    mocked_time.now.return_value = datetime1
    assert cur_time("%H:%M:%S") == str_time_s


@patch('trivial_tools.datetime_tools.text.datetime')
def test_cur_time_s(mocked_time, datetime1, str_time_s):
    """
    Получить строку с текущим временем с точностью до секунд
    """
    mocked_time.now.return_value = datetime1
    assert cur_time_s() == str_time_s


@patch('trivial_tools.datetime_tools.text.datetime')
def test_cur_time_ms(mocked_time, datetime1, str_time_ms):
    """
    Получить строку с текущим временем с точностью до микросекунд
    """
    mocked_time.now.return_value = datetime1
    assert cur_time_ms() == str_time_ms


def test_timestamp_to_str(timestamp):
    """
    Преобразовать timestamp в текстовую форму.
    """
    assert timestamp_to_str(timestamp) == '2019-12-11 11:26:48'


def test_date_to_text(date1, str_date1):
    """
    Преобразовать время в виде date в текстовую форму
    """
    assert date_to_text(date1) == str_date1


def test_datetime_to_text(datetime1, str_date1, datetime1_str_s):
    """
    Преобразовать время в виде datetime в текстовую форму
    """
    assert datetime_to_text(datetime1, "%Y-%m-%d") == str_date1
    assert datetime_to_text(datetime1, "%Y-%m-%d %H:%M:%S") == datetime1_str_s


def test_datetime_to_text_s(datetime1, datetime1_str_s):
    """
    Преобразовать время в виде datetime в текстовую форму
    """
    assert datetime_to_text_s(datetime1) == datetime1_str_s


def test_datetime_to_text_ms(datetime1, datetime1_str_ms):
    """
    Преобразовать время в виде datetime в текстовую форму
    """
    assert datetime_to_text_ms(datetime1) == datetime1_str_ms



# -*- coding: utf-8 -*-
"""

    Инструменты вычисления времени и дат

"""
# встроенные модули
import calendar
from typing import Union
from datetime import timedelta, datetime


def cast_seconds_to_timedelta(value: Union[timedelta, int, float]) -> timedelta:
    """Преобразовать число секунд в экземпляр timedelta (при возможности).

    :param value: число секунд или экземпляр timedelta
    :return: экземпляр timedelta

    >>> cast_seconds_to_timedelta(1)
    datetime.timedelta(seconds=1)
    """
    if isinstance(value, timedelta):
        return value
    return timedelta(seconds=float(value))


def days_in_month(year: int, month: int) -> int:
    """
    Получить число дней в месяце

    >>> days_in_month(2016, 1)
    31

    >>> days_in_month(2019, 4)
    30
    """
    _, length = calendar.monthrange(year, month)
    return length


def hours_in_month(year: int, month: int) -> int:
    """
    Получить число часов в месяце

    >>> hours_in_month(2016, 1)
    744

    >>> hours_in_month(2019, 4)
    720
    """
    result = days_in_month(year, month) * 24
    return result


def get_seconds_diff(old: datetime, new: datetime) -> float:
    """
    Получить число секунд между метками
    """
    return (new - old).total_seconds()


def get_seconds_diff_int(old: datetime, new: datetime) -> int:
    """
    Получить число секунд между метками
    """
    return int(get_seconds_diff(old, new))



# -*- coding: utf-8 -*-
"""

    Простые инструменты для работы с JSON-RPC

"""
# встроенные модули
from typing import Dict, Any, Union, Optional


def form_request(*, method: str, msg_id: Optional[Union[int, str]] = None,
                 **kwargs) -> Dict[str, Any]:
    """
    Собрать запрос к API
    """
    return {"jsonrpc": "2.0", "method": method, "params": {**kwargs}, "id": msg_id}


def result(output: Any, msg_id: Optional[Union[int, str]] = None) -> Dict[str, Any]:
    """
    Успешный результат
    """
    return {"jsonrpc": "2.0", "result": output, "id": msg_id}


def error(output: Any, msg_id: Optional[Union[int, str]] = None) -> Dict[str, Any]:
    """
    Неудачный результат
    """
    return {"jsonrpc": "2.0", "error": output, "id": msg_id}



# -*- coding: utf-8 -*-
"""

    Инструменты для работы с API

"""
# встроенные модули
from typing import Optional, Any

# сторонние модули
import requests
from loguru import logger

# модули проекта
from trivial_tools.special.special import fail
from trivial_tools.formatters.base import dict_as_args
from trivial_tools.runners.base import repeat_on_exceptions
from trivial_tools.json_rpc.basic_tools import form_request


@repeat_on_exceptions(repeats=5, case=Exception, delay=1)
def call_api(url: str, *, method: str, error_msg: str, debug: bool, **kwargs) -> Optional[Any]:
    """
    Выполнить POST запрос
    """
    payload = form_request(
        method=method,
        msg_id=kwargs.get('msg_id'),
        **kwargs
    )

    if debug:
        # запомнить, что мы запросили
        safe_kwargs = {
            key: value if key not in ('secret_key', 'fingerprint') else '***'
            for key, value in kwargs.items()
        }
        logger.debug(f'---> {method}({dict_as_args(safe_kwargs)})')

    r = requests.post(url, json=payload)

    if r.status_code == 200:
        body = r.json()

        if 'result' in body:
            result = body['result']

            if debug:
                # запомнить, что мы получили
                logger.debug(f'<--- {method}: <' + str(result)[0:100] + '>')

            return result

        else:
            post_mortem = body['error']
    else:
        post_mortem = r.content.decode('utf-8')

    logger.critical(post_mortem)
    fail(error_msg, reason=ConnectionError)
    return None



# -*- coding: utf-8 -*-
"""

    Класс с динамически генерируемыми атрибутами

"""
# модули проекта
from trivial_tools.special.special import fail
from trivial_tools.formatters.base import s_type


class Fluid:
    """
    Класс с динамически генерируемыми атрибутами
    """
    def __init__(self, *args, **kwargs):
        """
        Автоматическая инициация
        """
        if args:
            fail(f'{s_type(self)} подразумевает создание только из именованных атрибутов',
                 reason=ValueError)

        for key, value in kwargs.items():
            setattr(self, key, value)

    def to_dict(self) -> dict:
        """
        Преобразовать в словарь
        """
        return {key: value for key, value in vars(self).items()}

    def alter(self, **kwargs):
        """
        Получить аналогичный экземпляр, но с другими параметрами
        """
        values = {**self.to_dict(), **kwargs}
        return self.from_dict(values)

    @classmethod
    def from_dict(cls, parameters: dict):
        """
        Создать экземпляр из словаря
        """
        return cls(**parameters)

    def __str__(self) -> str:
        """
        Текстовое представление
        """
        fields = []
        for key in vars(self):
            value = getattr(self, key)
            fields.append((key, value))

        if len(fields) <= 3:
            strings = [f'{key}={value}' for key, value in fields]
            return f'{s_type(self)}(' + ', '.join(strings) + ')'

        longest = max([len(x[0]) for x in fields])
        offset = '    '

        result = []
        for i, (key, value) in enumerate(fields, start=1):
            if isinstance(value, (list, tuple)):
                if len(value) > 5:
                    # длинная последовательность
                    # 05. par_5 = [
                    #               1,
                    #               2,
                    string = ('{0:02d}. {1:' + str(longest) + '} = [').format(i, key)
                    result.append(offset + string)
                    for j, sub_element in enumerate(value, start=1):
                        if j == len(value):
                            result.append(offset + len(string) * ' ' + f'{sub_element}')
                        else:
                            result.append(offset + len(string) * ' ' + f'{sub_element},')
                    result.append(offset + (len(string) - 1) * ' ' + ']')
                else:
                    # короткая последовательность
                    # \t01. par_1 = [1, 2, 3]
                    string = ('{0:02d}. {1:' + str(longest) + '} = ').format(i, key)
                    result.append(offset + string + str(value))
            else:
                string = ('{0:02d}. {1:' + str(longest) + '} = {2}').format(i, key, value)
                result.append(offset + string)
        return s_type(self) + '(\n' + '\n'.join(result) + '\n)'

    def __repr__(self) -> str:
        """
        Текстовое представление
        """
        result = []
        for key in vars(self):
            value = getattr(self, key)
            result.append(f'{key}={value!r}')

        return f'{s_type(self)}(' + ', '.join(result) + ')'

    def __eq__(self, other) -> bool:
        """
        Равенство
        """
        if isinstance(other, type(self)):
            return self.to_dict() == other.to_dict()
        return False



# -*- coding: utf-8 -*-
"""

    Базовый класс для управления конфигурациями

"""
# встроенные модули
from contextlib import contextmanager

# модули проекта
from trivial_tools.classes.fluid import Fluid
from trivial_tools.config_handling import AbstractConfig
from trivial_tools.classes.conservator import Conservator


class BaseConfig(Fluid, Conservator, AbstractConfig):
    """
    Базовый класс конфигураций
    """
    @classmethod
    def from_json_if_need(cls, config_name: str, filename: str = 'config.json'):
        """
        Создать экземпляр только если он ещё не существует
        """
        if cls.has_instance():
            return cls.get_instance()
        return cls.from_json(config_name, filename)

    @classmethod
    @contextmanager
    def temporary_config(cls, **kwargs):
        """
        Контекстный менеджер для тестового экземпляра
        """
        if cls.has_instance():
            original = cls.get_instance()
            cls.unregister()
        else:
            original = None

        tmp_instance = cls(**kwargs)
        cls.register(tmp_instance)

        yield

        if original is None:
            cls.unregister()
        else:
            cls.replace(original)



# -*- coding: utf-8 -*-
"""

    Тесты вычислений

"""
# модули проекта
from trivial_tools.calculations.base_math import byte_count_to_si, byte_count_to_text


def test_byte_count_to_si_ru():
    """
    Проверить преобразование по основанию 1000, русский
    """
    assert byte_count_to_si(-2_000, lang='RU') == '-2.0 кБ'
    assert byte_count_to_si(-2_048, lang='RU') == '-2.0 кБ'
    assert byte_count_to_si(0, lang='RU') == '0 Б'
    assert byte_count_to_si(27, lang='RU') == '27 Б'
    assert byte_count_to_si(999, lang='RU') == '999 Б'
    assert byte_count_to_si(1_000, lang='RU') == '1.0 кБ'
    assert byte_count_to_si(1_023, lang='RU') == '1.0 кБ'
    assert byte_count_to_si(1_024, lang='RU') == '1.0 кБ'
    assert byte_count_to_si(1_728, lang='RU') == '1.7 кБ'
    assert byte_count_to_si(110_592, lang='RU') == '110.6 кБ'
    assert byte_count_to_si(7_077_888, lang='RU') == '7.1 МБ'
    assert byte_count_to_si(1_000_000, lang='RU') == '1.0 МБ'
    assert byte_count_to_si(452_984_832, lang='RU') == '453.0 МБ'
    assert byte_count_to_si(1_000_000_000, lang='RU') == '1.0 ГБ'
    assert byte_count_to_si(28_991_029_248, lang='RU') == '29.0 ГБ'
    assert byte_count_to_si(1_855_425_871_872, lang='RU') == '1.9 ТБ'
    assert byte_count_to_si(9_223_372_036_854_775_807, lang='RU') == '9.2 ЭБ'


def test_byte_count_to_si_en():
    """
    Проверить преобразование по основанию 1000, английский
    """
    assert byte_count_to_si(-2_000, lang='EN') == '-2.0 kB'
    assert byte_count_to_si(-2_048, lang='EN') == '-2.0 kB'
    assert byte_count_to_si(0, lang='EN') == '0 B'
    assert byte_count_to_si(27, lang='EN') == '27 B'
    assert byte_count_to_si(999, lang='EN') == '999 B'
    assert byte_count_to_si(1_000, lang='EN') == '1.0 kB'
    assert byte_count_to_si(1_023, lang='EN') == '1.0 kB'
    assert byte_count_to_si(1_024, lang='EN') == '1.0 kB'
    assert byte_count_to_si(1_728, lang='EN') == '1.7 kB'
    assert byte_count_to_si(110_592, lang='EN') == '110.6 kB'
    assert byte_count_to_si(7_077_888, lang='EN') == '7.1 MB'
    assert byte_count_to_si(1_000_000, lang='EN') == '1.0 MB'
    assert byte_count_to_si(452_984_832, lang='EN') == '453.0 MB'
    assert byte_count_to_si(1_000_000_000, lang='EN') == '1.0 GB'
    assert byte_count_to_si(28_991_029_248, lang='EN') == '29.0 GB'
    assert byte_count_to_si(1_855_425_871_872, lang='EN') == '1.9 TB'
    assert byte_count_to_si(9_223_372_036_854_775_807, lang='EN') == '9.2 EB'


def test_byte_count_to_text_ru():
    """
    Проверить преобразование по основанию 1024
    """
    assert byte_count_to_text(-2_000, lang='RU') == '-2.0 КиБ'
    assert byte_count_to_text(-2_048, lang='RU') == '-2.0 КиБ'
    assert byte_count_to_text(0, lang='RU') == '0 Б'
    assert byte_count_to_text(27, lang='RU') == '27 Б'
    assert byte_count_to_text(999, lang='RU') == '999 Б'
    assert byte_count_to_text(1_000, lang='RU') == '1000 Б'
    assert byte_count_to_text(1_023, lang='RU') == '1023 Б'
    assert byte_count_to_text(1_024, lang='RU') == '1.0 КиБ'
    assert byte_count_to_text(1_728, lang='RU') == '1.7 КиБ'
    assert byte_count_to_text(110_592, lang='RU') == '108.0 КиБ'
    assert byte_count_to_text(1_000_000, lang='RU') == '976.6 КиБ'
    assert byte_count_to_text(7_077_888, lang='RU') == '6.8 МиБ'
    assert byte_count_to_text(452_984_832, lang='RU') == '432.0 МиБ'
    assert byte_count_to_text(1_000_000_000, lang='RU') == '953.7 МиБ'
    assert byte_count_to_text(28_991_029_248, lang='RU') == '27.0 ГиБ'
    assert byte_count_to_text(1_855_425_871_872, lang='RU') == '1.7 ТиБ'
    assert byte_count_to_text(9_223_372_036_854_775_807, lang='RU') == '8.0 ЭиБ'


def test_byte_count_to_text_en():
    """
    Проверить преобразование по основанию 1024
    """
    assert byte_count_to_text(-2_000, lang='EN') == '-2.0 KiB'
    assert byte_count_to_text(-2_048, lang='EN') == '-2.0 KiB'
    assert byte_count_to_text(0, lang='EN') == '0 B'
    assert byte_count_to_text(27, lang='EN') == '27 B'
    assert byte_count_to_text(999, lang='EN') == '999 B'
    assert byte_count_to_text(1_000, lang='EN') == '1000 B'
    assert byte_count_to_text(1_023, lang='EN') == '1023 B'
    assert byte_count_to_text(1_024, lang='EN') == '1.0 KiB'
    assert byte_count_to_text(1_728, lang='EN') == '1.7 KiB'
    assert byte_count_to_text(110_592, lang='EN') == '108.0 KiB'
    assert byte_count_to_text(1_000_000, lang='EN') == '976.6 KiB'
    assert byte_count_to_text(7_077_888, lang='EN') == '6.8 MiB'
    assert byte_count_to_text(452_984_832, lang='EN') == '432.0 MiB'
    assert byte_count_to_text(1_000_000_000, lang='EN') == '953.7 MiB'
    assert byte_count_to_text(28_991_029_248, lang='EN') == '27.0 GiB'
    assert byte_count_to_text(1_855_425_871_872, lang='EN') == '1.7 TiB'
    assert byte_count_to_text(9_223_372_036_854_775_807, lang='EN') == '8.0 EiB'



# -*- coding: utf-8 -*-
"""

    Тесты простых операций преобразования времени

"""
# встроенные модули
from datetime import datetime

# сторонние модули
import pytest

# модули проекта
from trivial_tools.datetime_tools.basic import get_seconds_diff, get_seconds_diff_int


@pytest.fixture()
def time_1():
    return datetime(2019, 1, 1, 1, 1, 1, 530)


@pytest.fixture()
def time_2():
    return datetime(2019, 1, 1, 1, 1, 21, 740)


def test_get_seconds_diff(time_1, time_2):
    """
    Должен выдавать разницу в секундах
    """
    assert pytest.approx(get_seconds_diff(time_1, time_2)) == 20.00021


def test_get_seconds_diff_int(time_1, time_2):
    """
    Должен выдавать разницу в секундах
    """
    assert get_seconds_diff_int(time_1, time_2) == 20



# -*- coding: utf-8 -*-
"""

    Контейнер для вычисления скользящего среднего.
    В него можно постоянно можно добавлять элементы и быстро получать среднее

    Пример работы с контейнером:
    m = MovingAverage(maxlen=4)  # среднее равно 0
    m.push(1)   # среднее равно 1
    m.push(3)   # среднее равно 2
    m.push(5)   # среднее равно 3
    m.push(6)   # среднее равно 3,75
    m.push(8)   # среднее равно 5,5
    m.push(41)  # среднее равно 15
    m.push(9)   # среднее равно 16

"""
# встроенные модули
from typing import Union, Optional, Sequence, Any

# модули проекта
from trivial_tools.containers.class_moving_sum import MovingSum


class MovingAverage(MovingSum):
    """
    Контейнер для вычисления скользящего среднего.
    В него можно постоянно можно добавлять элементы и быстро получать среднее
    """
    __slots__ = ('_avg',)

    def __init__(self, source: Optional[Sequence] = None,
                 window: int = 0, sentinel: Any = object()):
        """
        Создание экземпляра

        :param source: исходная коллекция элементов, на базе которой надо собрать экземпляр
        :param window: максимальное киличество элементов (ширина окна вычисления)
        :param sentinel: элемент для заполнения пустых ячеек (можно добавить свой)
        """
        self._avg = 0.0
        super().__init__(source, window, sentinel)

    def push(self, value: Union[int, float]) -> None:
        """
        Добавить элемент в контейнер

        Мы выталкиваем предыдущий элемент и меняем сумму соответствующим образом. Причина
        для такого выполнения - избежать необходимости итерироваться по коллекции при
        попытке вычислить среднее значение

        :param value: новое число, которое соответствует сдвигу окна среднего на один элемент
        """
        super().push(value)
        self._avg = self._sum / len(self)

    def restore(self) -> None:
        """
        Сбросить параметры
        """
        self._avg = 0.0
        super().restore()

    @property
    def avg(self) -> float:
        """
        Среднее всех элементов
        """
        return self._avg

    def __setitem__(self, key: Union[int, slice], value: Union[int, float, Sequence]) -> Any:
        """
        Записать элемент по индексу.
        Обеспечивается обычный доступ к внутреннему хранилищу, просто со смещением индекса

        :param key: ключ индексации
        :param value: данные для записи (любой тип)
        """
        super().__setitem__(key, value)
        self._avg = self._sum / len(self)



# -*- coding: utf-8 -*-
"""

    Тесты одиночки

"""
# сторонние модули
import pytest

# модули проекта
from trivial_tools.classes.singleton_base import MetaSingleton


@pytest.fixture
def class_():
    class Test(metaclass=MetaSingleton):
        def __init__(self, x):
            self.x = x
    return Test


def test_creation(class_):
    """
    Создание экземпляра
    """
    t = class_(1)
    assert t
    assert t.x == 1


def test_creation2(class_):
    """
    Создание экземпляра второй раз, проверка работы фикстуры
    """
    t = class_(2)
    assert t
    assert t.x == 2


def test_creation3(class_):
    """
    Создание экземпляра, проверка собственно метакласса
    """
    t1 = class_(3)
    assert t1
    assert t1.x == 3

    t2 = class_(4)
    assert t2
    assert t2.x == 3


def test_get_instance(class_):
    """
    Создание экземпляра, проверка собственно метакласса
    """
    t1 = class_(4)
    assert t1
    assert t1.x == 4

    t2 = class_.get_instance(t1)
    assert t1 is t2

    t3 = class_.get_instance(class_)
    assert t1 is t3


def test_get_instance_fail(class_):
    """
    Создание экземпляра, проверка собственно метакласса
    """
    with pytest.raises(KeyError):
        MetaSingleton.get_instance(1)



# -*- coding: utf-8 -*-
"""

    Регистратор методов для JSON-RPC.
    Используется для централизованного управления всеми методами приложения.

"""
# встроенные модули
import json
from inspect import signature
from types import FunctionType
from typing import Optional, List, Callable, Dict, Sequence, Tuple, Any, Union

# сторонние модули
from trivial_tools.formatters.base import s_type

JSON_RPC = Dict[str, Any]


OK: int = 200
BAD_REQUEST: int = 400
NOT_AUTHORIZED: int = 401
METHOD_NOT_ALLOWED: int = 405
NOT_ACCEPTABLE: int = 406
INTERNAL_ERROR: int = 500


class JSONRPCMethodMaster:
    """
    Регистратор методов для JSON-RPC.
    Используется для централизованного управления всеми методами приложения.
    """
    def __init__(self, ignores: Sequence[str] = ('secret_key',)):
        """
        Инициация
        """
        self._ignores = ['return'] + list(ignores)
        self._methods: Dict[str, Callable] = {}

    def __contains__(self, item):
        """
        Проверить, имеется ли этот метод в записях.
        """
        return item in self._methods

    def register_method(self, func: FunctionType) -> FunctionType:
        """
        Зарегистрировать метод.
        """
        self._methods[func.__name__] = func
        return func

    def unknown_method(self, method_name: str) -> bool:
        """
        Этот метод нам неизвестен?
        """
        return method_name not in self._methods

    def check_signature(self, body: Union[JSON_RPC, List[JSON_RPC]]):
        """
        Проверка полученных параметров на соответствие сигнатуре метода
        """
        response = []
        return_code = NOT_ACCEPTABLE

        if isinstance(body, dict):
            error_message = self._check_signature(body)
            if error_message:
                return {'code': -32602, 'message': error_message}, return_code
            else:
                return body, OK

        else:
            for sub_body in body:
                error_message = self._check_signature(sub_body)
                if error_message:
                    response.append(({'code': -32602, 'message': error_message}, return_code))
                else:
                    response.append(({'code': -32602, 'message': error_message}, OK))

        return response, return_code

    def _check_signature(self, body: JSON_RPC) -> str:
        """
        Проверка полученных параметров на соответствие сигнатуре метода
        """
        method_name = body.get('method')
        parameters = body.get('params').keys()

        method: FunctionType = self._methods[method_name]

        sig = signature(method)

        valid_keys = {x for x in sig.parameters if x not in self._ignores}
        other_keys = {x for x in parameters if x not in self._ignores}

        response = 'Расхождение в аргументах метода: '

        if valid_keys != other_keys:
            insufficient = ', '.join(valid_keys - other_keys)
            excess = ', '.join(other_keys - valid_keys)

            if insufficient:
                response += 'не хватает аргументов ' + insufficient

                if excess:
                    response += ', '

            if excess:
                response += 'лишние аргументы ' + excess

            return response
        return ''

    def get_method(self, method_name: str) -> Optional[Callable]:
        """
        Получить метод
        """
        return self._methods.get(method_name)

    def get_method_names(self) -> List[str]:
        """
        Получить имена доступных методов
        """
        return sorted(self._methods.keys())

    @staticmethod
    def extract_json(request) -> Tuple[Union[JSON_RPC, List[JSON_RPC]], int]:
        """
        Безопасное извлечение JSON из запроса
        """
        try:
            request_body = request.json()
            return_code = OK

        except json.JSONDecodeError as err:
            request_body = {'code': -32700, 'message': f'Неправильный формат запроса, {err}'}
            return_code = BAD_REQUEST

        return request_body, return_code

    def check_request(self, body: Union[JSON_RPC, List[JSON_RPC]])\
            -> Tuple[Union[JSON_RPC, List[JSON_RPC]], int]:
        """
        Проверить на предмет соответствия JSON RPC 2.0
        """
        return_code = BAD_REQUEST

        if isinstance(body, dict):
            response, return_code = self.check_dict(body)

        elif isinstance(body, list):
            response = []
            for sub_body in body:
                checked_body, return_code = self.check_dict(sub_body)

                if return_code != OK:
                    break

                response.append(checked_body)
        else:
            response = {'code': -32700, 'message': f'Неправильный формат запроса, {s_type(body)}'}

        return response, return_code

    def check_dict(self, body: JSON_RPC) -> Tuple[JSON_RPC, int]:
        """
        Проверить словарь запроса
        """
        return_code = BAD_REQUEST

        version = body.get('jsonrpc')
        method = body.get('method')
        params = body.get('params')

        if version != '2.0':
            return {'code': -32600,
                    'message': 'Поддерживается только JSON-RPC 2.0.'}, return_code

        if method is None:
            return {'code': -32601,
                    'message': 'Не указан метод запроса.'}, return_code

        elif method not in self._methods:
            return {'code': -32601,
                    'message': f'Неизвестный метод: {method}.'}, METHOD_NOT_ALLOWED

        if params is None:
            return {'code': -32602,
                    'message': 'Не указаны аргументы вызова.'}, return_code

        elif params.get('secret_key') is None:
            return {'code': -32602,
                    'message': 'Не предоставлен ключ авторизации.'}, NOT_AUTHORIZED

        elif isinstance(params, list):
            return {'code': -32602,
                    'message': 'Передача аргументов по позиции не поддерживается.'}, NOT_ACCEPTABLE

        elif not isinstance(params, dict):
            return {'code': -32602,
                    'message': f'Неправильно оформлены аргументы: {params}'}, NOT_ACCEPTABLE

        return body, OK

    @staticmethod
    def authorize(body: JSON_RPC, check_func: Callable) -> Tuple[JSON_RPC, int]:
        """
        Проверить авторизацию внешней функцией
        """
        checked_body, return_code = check_func(body)
        if return_code != OK:
            return {'code': -32602,
                    'message': f'Отказано в доступе.'}, NOT_AUTHORIZED
        return checked_body, OK

    def execute(self, body: JSON_RPC) -> Any:
        """
        Исполнить запрос
        """
        method_name = body.get('method')
        params = body.get('params', {})
        method = self.get_method(method_name)

        # секуретный ключ не нужен внутри методов
        params.pop('secret_key', None)

        response = method(**params)
        return response
